'''
Test GENIE filters
'''
import datetime
import os
import sys
from unittest.mock import patch

import pandas as pd

from genie.process_functions import seqDateFilter
from genie import database_to_staging
from genie.database_to_staging import (seq_assay_id_filter,
                                       redact_phi,
                                       no_genepanel_filter,
                                       _to_redact_interval,
                                       _redact_year,
                                       _to_redact_difference)
from genie.consortium_to_public import commonVariantFilter

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(SCRIPT_DIR, "../../genie"))


# syn = synapseclient.login()
def test_seqassayfilter():
    # SEQ ASSAY ID filter, only seq assay ids with more than
    sagetest = ["SAGE-TEST-1"]*51
    sagetest.extend(["SAGE-TEST-2"]*10)

    clinicalDf = pd.DataFrame(sagetest)
    clinicalDf.rename(columns={0: "SEQ_ASSAY_ID"}, inplace=True)
    clinicalDf['SAMPLE_ID'] = range(0, len(sagetest))
    # Make sure that only samples of seq assay ids that have 50 or
    # more samples are returned
    samples = seq_assay_id_filter(clinicalDf).reset_index()
    # Must make sure that seq assays have the same index to do comparisons
    del samples['index']
    samples = samples['SAMPLE_ID']
    assert all(samples == pd.Series(range(51, 61)))


def test_seqdatefilter():
    # Check that seq date filter works correctly
    SEQ_DATE = "Jan-2018"
    processingDate = datetime.datetime.strptime(SEQ_DATE, '%b-%Y')
    clinicalDf = pd.DataFrame([
        'SAGE-TEST-1', 'SAGE-TEST-2', 'SAGE-TEST-3',
        'SAGE-TEST-4', 'SAGE-TEST-5', 'SAGE-TEST-6', 'SAGE-TEST-7'])
    clinicalDf.rename(columns={0: "SAMPLE_ID"}, inplace=True)
    expected = [
        'SAGE-TEST-1', 'SAGE-TEST-2', 'SAGE-TEST-3',
        'SAGE-TEST-4', 'SAGE-TEST-5', 'SAGE-TEST-6']
    clinicalDf['SEQ_DATE'] = [
        "Jan-2018", "Apr-2018", "Oct-2017",
        "Jul-2017", "Apr-2017", "Jan-2017", "Oct-2016"]
    samples = seqDateFilter(clinicalDf, processingDate, 366)
    assert all(samples == expected)

    # Half year filter
    samples = seqDateFilter(clinicalDf, processingDate, 184)
    expected = ['SAGE-TEST-1', 'SAGE-TEST-2', 'SAGE-TEST-3', 'SAGE-TEST-4']
    assert all(samples == expected)

    # Test leap year
    SEQ_DATE = "Jan-2017"
    processingDate = datetime.datetime.strptime(SEQ_DATE, '%b-%Y')
    clinicalDf = pd.DataFrame([
        'SAGE-TEST-1', 'SAGE-TEST-2', 'SAGE-TEST-3',
        'SAGE-TEST-4', 'SAGE-TEST-5', 'SAGE-TEST-6', 'SAGE-TEST-7'])
    clinicalDf.rename(columns={0: "SAMPLE_ID"}, inplace=True)
    clinicalDf['SEQ_DATE'] = [
        "Jan-2017", "Apr-2017", "Oct-2016",
        "Jul-2016", "Apr-2016", "Jan-2016", "Oct-2015"]
    expected = [
        'SAGE-TEST-1', 'SAGE-TEST-2', 'SAGE-TEST-3',
        'SAGE-TEST-4', 'SAGE-TEST-5', 'SAGE-TEST-6']
    samples = seqDateFilter(clinicalDf, processingDate, 366)
    assert all(samples == expected)

    # Check half year release date during leap year
    SEQ_DATE = "Jul-2016"
    processingDate = datetime.datetime.strptime(SEQ_DATE, '%b-%Y')
    samples = seqDateFilter(clinicalDf, processingDate, 184)
    assert all(samples == expected)


def test__to_redact_interval():
    """Tests the correct boolean vectors are returned for phi and pediatric
    redaction"""
    values = pd.Series([32850, 32485, 6570, 6569, '<foo', '>testing'])
    to_redact, to_redact_peds = _to_redact_interval(values)
    expected_redact = [True, False, False, False, False, True]
    expected_redact_peds = [False, False, False, True, True, False]
    assert to_redact.to_list() == expected_redact
    assert to_redact_peds.to_list() == expected_redact_peds


def test__redact_year():
    """Tests redaction of birth year based on < and >"""
    values = pd.Series([1923, 2003, '<foo', '>testing'])
    redacted = _redact_year(values)
    expected_redact = [1923, 2003, 'withheld', 'cannotReleaseHIPAA']
    assert redacted.to_list() == expected_redact


def test___to_redact_difference():
    """Tests if a difference between two year columns is >89, redact"""
    year1 = pd.Series([1923, 2000, float('nan')])
    year2 = pd.Series([1926, 2100, 2000])
    redacted = _to_redact_difference(year1, year2)
    expected_redact = [False, True, False]
    assert redacted.to_list() == expected_redact


def test_redact_phi():
    """Redacts PHI interval and years"""
    return_bools = ([True, False, False],
                    [False, False, True])
    clinicaldf = pd.DataFrame(['SAGE-TEST-1', 'SAGE-TEST-2',
                               'SAGE-TEST-3'])
    clinicaldf.rename(columns={0: "SAMPLE_ID"}, inplace=True)
    clinicaldf['AGE_AT_SEQ_REPORT'] = [32850, 32485, 6570]
    clinicaldf['BIRTH_YEAR'] = [1900, '<1900', 1902]
    # These are the years that are returned by the _redact_year
    # Use these against YEAR_CONTACT and YEAR_DEATH to calculate
    # expected_birth
    return_year = pd.Series([2000, 1903, 1904])
    clinicaldf['YEAR_CONTACT'] = [2100, 1904, float('nan')]
    clinicaldf['YEAR_DEATH'] = [2001, float('nan'), 2000]

    expected_age = pd.Series(['>32485', 32485, '<6570'])
    expected_birth = pd.Series(['cannotReleaseHIPAA', 1903,
                                'cannotReleaseHIPAA'])
    with patch.object(database_to_staging, "_to_redact_interval",
                      return_value=return_bools),\
         patch.object(database_to_staging, "_redact_year",
                      return_value=return_year):
        redacted_clin = redact_phi(clinicaldf,
                                   interval_cols_to_redact=['AGE_AT_SEQ_REPORT'])
        assert all(redacted_clin['AGE_AT_SEQ_REPORT'] == expected_age)
        assert all(redacted_clin['BIRTH_YEAR'] == expected_birth)

# def test_MAFinBED():
#   syn = mock.create_autospec(synapseclient.Synapse)
#   # MAF in BED filter (Make sure that the only Hugo symbol that
    # passes through this filter is the ERRFI1 gene)
#   databaseSynIdMapping = syn.tableQuery('select * from syn11600968')
#   databaseSynIdMappingDf = databaseSynIdMapping.asDataFrame()
    # CENTER_MAPPING = syn.tableQuery(
    #     'SELECT * FROM syn11601248 where stagingSynId is not null')
#   CENTER_MAPPING_DF = CENTER_MAPPING.asDataFrame()
#   mafPath = os.path.join(SCRIPT_DIR,"testingmaf.txt")
    # temp = runMAFinBED(
    #     syn, mafPath, CENTER_MAPPING_DF, databaseSynIdMappingDf, test=True)
#   maf = pd.read_csv(mafPath,sep="\t")
#   assert all(maf['Hugo_Symbol'][maf['inBED'] ==  True].unique() == "ERRFI1")
#   os.unlink(mafPath)

# def test_MutationInCis():
#   def createMockTable(dataframe):
#       table = mock.create_autospec(synapseclient.table.CsvFileTable)
#       table.asDataFrame.return_value= dataframe
#       return(table)

#   def table_query_results(*args):
#       return(table_query_results_map[args])

#   databaseMapping = pd.DataFrame(dict(Database=['mutationsInCis'],
#                                       Id=['syn7765462']))

#   centerMapping = pd.DataFrame(dict(center=['SAGE'],
#                                     inputSynId=['syn11601335'],
#                                     stagingSynId=['syn11601337']))

#   mutCisDf = pd.DataFrame(dict(Flag=['mutationsInCis'],
#                                  Center=['syn7765462'],
#                                  Tumor_Sample_Barcode=["GENIE-SAGE-ID1-1"],
#                                  Variant_Classification=["Nonsense_Mutation"],
#                                  Hugo_Symbol=["AKT1"],
#                                  HGVSp_Short=["p.1234"],
#                                  Chromosome=["1"],
#                                  Start_Position=[324234],
#                                  Reference_Allele=["AGC"],
#                                  Tumor_Seq_Allele2=["GCCCT"],
#                                  t_alt_count_num=[3],
#                                  t_depth=[234]))

#   #This is the gene positions that all bed
#   dataframe will be processed against
    # table_query_results_map = {
    #     ('select * from syn11600968',):
    #         createMockTable(databaseMapping),
    #     ('SELECT * FROM syn11601248 where staginsgSynId is not null',):
    #         createMockTable(centerMapping),
    #     ("select * from syn7765462 where Center = 'SAGE'",):
    #         createMockTable(mutCisDf)
    # }

#   syn = mock.create_autospec(synapseclient.Synapse)
#   syn.tableQuery.side_effect=table_query_results
#   # Mutation in cis filter check (There is one TOSS sample,
#   it should be GENIE-TEST-0-1)
    # remove_samples = mutation_in_cis_filter(
    #     syn, True, None, None, "syn11601206", None)
#   assert  all(remove_samples == "GENIE-TEST-0-1")


def test_commonvariantfilter():
    '''
    Test common variant filter, make sure none
    of the common_variant rows are kept
    '''
    mutationDf = pd.DataFrame([
        "common_variant", "test1", "asdfasd:common_variant",
        "fo;common_variant", "test2", "test3"])
    mutationDf.rename(columns={0: "FILTER"}, inplace=True)
    maf = commonVariantFilter(mutationDf)
    expected = ["test1", "test2", "test3"]
    assert all(maf['FILTER'] == expected)


def test_no_genepanel_filter():
    '''
    Tests the filter to remove samples
    with no bed files
    '''
    sagetest = ["SAGE-TEST-1"]*51
    beddf = pd.DataFrame(sagetest)
    beddf.rename(columns={0: "SEQ_ASSAY_ID"}, inplace=True)
    sagetest.extend(["SAGE-TEST-2"]*10)

    clinicaldf = pd.DataFrame(sagetest)
    clinicaldf.rename(columns={0: "SEQ_ASSAY_ID"}, inplace=True)
    clinicaldf['SAMPLE_ID'] = sagetest

    remove_samples = no_genepanel_filter(clinicaldf, beddf)
    assert all(remove_samples == ["SAGE-TEST-2"]*10)



"""Tests validate.py"""
from unittest import mock
from unittest.mock import Mock, patch

import pandas as pd
import pytest
import synapseclient
from synapseclient.core.exceptions import SynapseHTTPError

from genie import validate, process_functions, example_filetype_format

CENTER = "SAGE"
syn = mock.create_autospec(synapseclient.Synapse)

CNA_ENT = synapseclient.File(name="data_CNA_SAGE.txt",
                             path="data_CNA_SAGE.txt",
                             parentId="syn12345")
CLIN_ENT = synapseclient.File(name="data_clinical_supp_SAGE.txt",
                              path="data_clinical_supp_SAGE.txt",
                              parentId="syn12345")
SAMPLE_ENT = synapseclient.File(name="data_clinical_supp_sample_SAGE.txt",
                                path="data_clinical_supp_sample_SAGE.txt",
                                parentId="syn12345")
PATIENT_ENT = synapseclient.File(name="data_clinical_supp_patient_SAGE.txt",
                                 path="data_clinical_supp_patient_SAGE.txt",
                                 parentId="syn12345")
WRONG_NAME_ENT = synapseclient.File(name="wrong.txt",
                                    path="data_clinical_supp_SAGE.txt",
                                    parentId="syn12345")


class FileFormat(example_filetype_format.FileTypeFormat):
    _fileType = "clinical"


def test_perfect_determine_filetype():
    """
    Tests determining of file type through filenames
    Parameters are passed in from filename_fileformat_map
    """
    filetype = "clincial"
    ent_list = [SAMPLE_ENT]
    with patch.object(FileFormat, "validateFilename",
                      return_value=filetype):
        validator = validate.GenieValidationHelper(
            syn, None, CENTER, ent_list,
            format_registry={filetype: FileFormat}
        )
        assert validator.determine_filetype() == filetype


def test_wrongfilename_noerror_determine_filetype():
    '''
    Tests None is passed back when wrong filename is passed
    when raise_error flag is False
    '''
    ent_list = [WRONG_NAME_ENT]
    with patch.object(FileFormat, "validateFilename",
                      side_effect=AssertionError):
        validator = validate.GenieValidationHelper(
            syn, project_id=None,
            center=CENTER, entitylist=ent_list,
            format_registry={"wrong": FileFormat})
        assert validator.file_type is None


def test_valid_collect_errors_and_warnings():
    '''
    Tests if no error and warning strings are passed that
    returned valid and message is correct
    '''
    message = validate.collect_errors_and_warnings('', '')
    assert message == "YOUR FILE IS VALIDATED!\n"


def test_invalid_collect_errors_and_warnings():
    """
    Tests if error and warnings strings are passed that
    returned valid and message is correct
    """
    message = validate.collect_errors_and_warnings("error\nnow",
                                                   'warning\nnow')
    assert message == (
        "----------------ERRORS----------------\n"
        "error\nnow"
        "-------------WARNINGS-------------\n"
        'warning\nnow')


def test_warning_collect_errors_and_warnings():
    """
    Tests if no error but warnings strings are passed that
    returned valid and message is correct
    """
    message = \
        validate.collect_errors_and_warnings('', 'warning\nnow')
    assert message == (
        "YOUR FILE IS VALIDATED!\n"
        "-------------WARNINGS-------------\n"
        'warning\nnow')


def test_valid_validate_single_file():
    """
    Tests that all the functions are run in validate single
    file workflow and all the right things are returned
    """
    entitylist = [CLIN_ENT]
    error_string = ''
    warning_string = ''
    expected_valid = True
    expected_message = "valid message here!"
    expected_filetype = "clinical"
    project_ent = Mock(id='syn1234')
    with patch.object(syn, "get", return_value=project_ent),\
         patch.object(validate.GenieValidationHelper,
                      "determine_filetype",
                      return_value=expected_filetype) as mock_determine_ftype,\
         patch.object(FileFormat, "validate",
                      return_value=(expected_valid, error_string,
                                    warning_string)) as mock_genie_class,\
         patch.object(validate, "collect_errors_and_warnings",
                      return_value=expected_message) as mock_determine:
        validator = validate.GenieValidationHelper(syn, project_id="syn1234",
                                                   center=CENTER,
                                                   entitylist=entitylist,
                                                   format_registry={'clinical': FileFormat})
        valid, message = validator.validate_single_file(oncotree_link=None,
                                                        nosymbol_check=False)

        assert valid == expected_valid
        assert message == expected_message
        assert validator.file_type == expected_filetype

        mock_determine_ftype.assert_called_once_with()

        mock_genie_class.assert_called_once_with(filePathList=[CLIN_ENT.path],
                                                 oncotree_link=None,
                                                 nosymbol_check=False,
                                                 project_id='syn1234')

        mock_determine.assert_called_once_with(error_string, warning_string)


def test_filetype_validate_single_file():
    """
    Tests that if filetype is passed in that an error is thrown
    if it is an incorrect filetype
    """
    entitylist = [WRONG_NAME_ENT]
    expected_error = ("----------------ERRORS----------------\n"
                      "Your filename is incorrect! Please change your "
                      "filename before you run the validator or specify "
                      "--filetype if you are running the validator locally")

    with patch.object(FileFormat, "validateFilename",
                      side_effect=AssertionError):
        validator = validate.GenieValidationHelper(
            syn, None, CENTER, entitylist,
            format_registry={'wrong': FileFormat}
        )

        valid, message = validator.validate_single_file()
        assert message == expected_error
        assert not valid


def test_wrongfiletype_validate_single_file():
    """
    Tests that if there is no filetype for the filename passed
    in, an error is thrown
    """
    entitylist = [WRONG_NAME_ENT]
    expected_error = ('----------------ERRORS----------------\n'
                      'Your filename is incorrect! Please change your '
                      'filename before you run the validator or specify '
                      '--filetype if you are running the validator locally')

    with patch.object(validate.GenieValidationHelper,
                      "determine_filetype",
                      return_value=None) as mock_determine_filetype:
        validator = validate.GenieValidationHelper(
            syn=syn, project_id=None, center=CENTER,
            entitylist=entitylist,
            format_registry={'wrong': Mock()})
        valid, message = validator.validate_single_file()

        assert message == expected_error
        assert not valid
        mock_determine_filetype.assert_called_once_with()


def test_nopermission__check_parentid_permission_container():
    """Throws error if no permissions to access"""
    parentid = "syn123"
    with patch.object(syn, "get", side_effect=SynapseHTTPError),\
         pytest.raises(ValueError,
                       match="Provided Synapse id must be your input folder "
                             "Synapse id or a Synapse Id of a folder inside "
                             "your input directory"):
        validate._check_parentid_permission_container(syn, parentid)


def test_notcontainer__check_parentid_permission_container():
    """Throws error if input if synid of file"""
    parentid = "syn123"
    file_ent = synapseclient.File("foo", parentId=parentid)
    with patch.object(syn, "get", return_value=file_ent),\
         pytest.raises(ValueError,
                       match="Provided Synapse id must be your input folder "
                             "Synapse id or a Synapse Id of a folder inside "
                             "your input directory"):
        validate._check_parentid_permission_container(syn, parentid)


def test_valid__check_parentid_permission_container():
    """
    Test that parentid specified is a container and have permissions to access
    """
    parentid = "syn123"
    folder_ent = synapseclient.Folder("foo", parentId=parentid)
    with patch.object(syn, "get", return_value=folder_ent):
        validate._check_parentid_permission_container(syn, parentid)


def test_valid__check_center_input():
    center = "FOO"
    center_list = ["FOO", "WOW"]
    validate._check_center_input(center, center_list)


def test_invalid__check_center_input():
    center = "BARFOO"
    center_list = ["FOO", "WOW"]
    with pytest.raises(ValueError,
                       match="Must specify one of these "
                             "centers: {}".format(", ".join(center_list))):
        validate._check_center_input(center, center_list)


ONCOTREE_ENT = 'syn222'


class argparser:
    oncotree_link = "link"
    parentid = None
    filetype = None
    project_id = None
    center = "try"
    filepath = "path.csv"
    nosymbol_check = False
    format_registry_packages = ["genie"]
    project_id = "syn1234"

    def asDataFrame(self):
        database_dict = {"Database": ["centerMapping", 'oncotreeLink'],
                         "Id": ["syn123", ONCOTREE_ENT],
                         "center": ["try", 'foo']}
        databasetosynid_mappingdf = pd.DataFrame(database_dict)
        return databasetosynid_mappingdf


def test_notnone_get_oncotree_link():
    """Test link passed in by user is used"""
    arg = argparser()
    url = "https://www.synapse.org"
    link = validate._get_oncotreelink(syn, arg.asDataFrame(),
                                      oncotree_link=url)
    assert link == url


def test_none__getoncotreelink():
    """Test oncotree link is gotten"""
    arg = argparser()
    url = "https://www.synapse.org"
    link = synapseclient.File("foo", parentId="foo", externalURL=url)
    with patch.object(syn, "get", return_value=link) as patch_synget:
        oncolink = validate._get_oncotreelink(syn, arg.asDataFrame())
        patch_synget.assert_called_once_with(ONCOTREE_ENT)
        assert oncolink == url


def test_valid__upload_to_synapse():
    """
    Test upload of file to synapse under right conditions
    """
    ent = synapseclient.File(id="syn123", parentId="syn222")
    with patch.object(syn, "store", return_value=ent) as patch_synstore:
        validate._upload_to_synapse(syn, ['foo'], True, parentid="syn123")
        patch_synstore.assert_called_once_with(
            synapseclient.File('foo', parent="syn123"))


def test_perform_validate():
    """Make sure all functions are called"""
    arg = argparser()
    valid = True
    with patch.object(validate,
                      "_check_parentid_permission_container",
                      return_value=None) as patch_check_parentid,\
         patch.object(process_functions, "get_synid_database_mappingdf",
                      return_value=arg.asDataFrame()) as patch_getdb,\
         patch.object(syn, "tableQuery",
                      return_value=arg) as patch_syn_tablequery,\
         patch.object(validate, "_check_center_input") as patch_check_center,\
         patch.object(validate, "_get_oncotreelink") as patch_get_onco,\
         patch.object(validate.GenieValidationHelper,
                      "validate_single_file",
                      return_value=(valid, 'foo')) as patch_validate,\
         patch.object(validate, "_upload_to_synapse") as patch_syn_upload:
        validate._perform_validate(syn, arg)
        patch_check_parentid.assert_called_once_with(syn, arg.parentid)
        patch_getdb.assert_called_once_with(syn, project_id=arg.project_id)
        patch_syn_tablequery.assert_called_once_with('select * from syn123')
        patch_check_center.assert_called_once_with(arg.center, ["try", "foo"])
        patch_get_onco.assert_called_once()
        patch_validate.assert_called_once_with(oncotree_link=arg.oncotree_link,
                                               nosymbol_check=arg.nosymbol_check,
                                               project_id=arg.project_id)
        patch_syn_upload.assert_called_once_with(
            syn, arg.filepath, valid, parentid=arg.parentid)



"""Initialize GENIE registry"""
# Import logging last to not take in synapseclient logging
import logging

from . import bed
from . import vcf
from . import bedSP
from . import workflow
from . import clinical
from . import seg
from . import cbs
from . import maf
from . import mafSP
from . import clinicalSP
from . import cna
from . import fusions
from . import sampleRetraction
from . import patientRetraction
from . import mutationsInCis
from . import assay

from genie.__version__ import __version__

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logging.getLogger("keyring").setLevel(logging.WARNING)



"""Test GENIE Bed class"""
import tempfile
import shutil
from unittest import mock
from unittest.mock import patch

import pandas as pd
from pandas.testing import assert_frame_equal
import pytest
import synapseclient

import genie_registry.bed
from genie_registry.bed import bed
from genie_registry.bedSP import bedSP

if not shutil.which('bedtools'):
    pytest.skip("bedtools is not found, skipping bed tests", 
                allow_module_level=True)

GENE_GTF_TEXT = '2\tprotein_coding\tgene\t69688532\t69901481\t.\t-\t.\tgene_id "ENSG00000115977"; gene_name "AAK1"; gene_source "ensembl_havana"; gene_biotype "protein_coding";\n9\tprotein_coding\tgene\t99401859\t99417585\t.\t-\t.\tgene_id "ENSG00000158122"; gene_name "AAED1"; gene_source "ensembl_havana"; gene_biotype "protein_coding";\n12\tprotein_coding\tgene\t53701240\t53718648\t.\t-\t.\tgene_id "ENSG00000094914"; gene_name "AAAS"; gene_source "ensembl_havana"; gene_biotype "protein_coding";\n19\tprotein_coding\tgene\t44047192\t44084625\t.\t-\t.\tgene_id "ENSG00000073050"; gene_name "XRCC1"; gene_source "ensembl_havana"; gene_biotype "protein_coding";\n19\tprotein_coding\tgene\t44080952\t44088116\t.\t+\t.\tgene_id "ENSG00000234465"; gene_name "PINLYP"; gene_source "ensembl_havana"; gene_biotype "protein_coding";'
EXON_GTF_TEXT = '2\tprocessed_transcript\texon\t69688432\t69689532\t.\t-\t.\tgene_id "ENSG00000115977"; transcript_id "ENST00000492192"; exon_number "2"; gene_name "AAK1"; gene_source "ensembl_havana"; gene_biotype "protein_coding"; transcript_name "AAK1-009"; transcript_source "havana"; exon_id "ENSE00001882560";\n9\tprotein_coding\texon\t99416987\t99417030\t.\t-\t.\tgene_id "ENSG00000158122"; transcript_id "ENST00000411939"; exon_number "1"; gene_name "AAED1"; gene_source "ensembl_havana"; gene_biotype "protein_coding"; transcript_name "AAED1-003"; transcript_source "havana"; exon_id "ENSE00001768346"; tag "cds_start_NF"; tag "mRNA_start_NF";\n12\tretained_intron\texon\t53702509\t53702599\t.\t-\t.\tgene_id "ENSG00000094914"; transcript_id "ENST00000550033"; exon_number "4"; gene_name "AAAS"; gene_source "ensembl_havana"; gene_biotype "protein_coding"; transcript_name "AAAS-019"; transcript_source "havana"; exon_id "ENSE00003694270";\n19\tprotein_coding\texon\t44084517\t44084625\t.\t-\t.\tgene_id "ENSG00000073050"; transcript_id "ENST00000598165"; exon_number "1"; gene_name "XRCC1"; gene_source "ensembl_havana"; gene_biotype "protein_coding"; transcript_name "XRCC1-008"; transcript_source "havana"; exon_id "ENSE00003137784"; tag "cds_end_NF"; tag "mRNA_end_NF";\n19\tprotein_coding\texon\t44084696\t44084739\t.\t+\t.\tgene_id "ENSG00000234465"; transcript_id "ENST00000562255"; exon_number "1"; gene_name "PINLYP"; gene_source "ensembl_havana"; gene_biotype "protein_coding"; transcript_name "PINLYP-001"; transcript_source "havana"; tag "CCDS"; ccds_id "CCDS58667"; exon_id "ENSE00002599477";'
GENE_TEMP = tempfile.NamedTemporaryFile()
EXON_TEMP = tempfile.NamedTemporaryFile()

with open(GENE_TEMP.name, "w") as gene:
    gene.write(GENE_GTF_TEXT)
with open(EXON_TEMP.name, "w") as exon:
    exon.write(EXON_GTF_TEXT)

def create_mock_table(dataframe):
    table = mock.create_autospec(synapseclient.table.CsvFileTable)
    table.asDataFrame.return_value = dataframe
    return(table)


def table_query_results(*args):
    return(table_query_results_map[args])


symbols = pd.DataFrame(dict(
    hgnc_symbol=['AAK1', 'AAED1', 'AAAS', 'PINLYP', 'XRCC1'],
    chromosome_name=['2', '9', '12', '19', '19'],
    start_position=[69688532, 99401859, 53701240, 44080952, 44047192],
    end_position=[69901481, 99417585, 53718648, 44088116, 44084625]))

# This is the gene positions that all bed dataframe will be processed against
table_query_results_map = {
    ("SELECT * FROM syn11806563",): create_mock_table(symbols), }

syn = mock.create_autospec(synapseclient.Synapse)
syn.tableQuery.side_effect = table_query_results

seq_assay_id = "SAGE-Test"
new_path = "new.bed"
parentid = "synTest"
bed_class = bed(syn, "SAGE")
bedsp_class = bedSP(syn, "SAGE")


def test_perfect___process():
    """Process perfect bed file"""
    expected_beddf = pd.DataFrame(dict(
        Chromosome=['2', '9', '12', '19', '19'],
        Start_Position=[69688533, 99401860, 53701241, 44084466, 44084466],
        End_Position=[69901480, 99417584, 53718647, 44084638, 44084638],
        Hugo_Symbol=['AAK1', 'AAED1', 'AAAS', 'XRCC1', 'PINLYP'],
        includeInPanel=[True, True, True, True, True],
        clinicalReported=[True, True, False, False, True],
        ID=['AAK1', 'AAED1', 'AAAS', 'XRCC1', 'foo'],
        SEQ_ASSAY_ID=['SAGE-TEST', 'SAGE-TEST',
                      'SAGE-TEST', 'SAGE-TEST', 'SAGE-TEST'],
        Feature_Type=['exon', 'exon', 'exon', 'exon', 'exon'],
        CENTER=['SAGE', 'SAGE', 'SAGE', 'SAGE', 'SAGE']))

    expected_beddf.sort_values("ID", inplace=True)
    expected_beddf.reset_index(drop=True, inplace=True)
    beddf = pd.DataFrame({
        0: ['2', '9', '12', '19', '19'],
        1: [69688533, 99401860, 53701241, 44084466, 44084466],
        2: [69901480, 99417584, 53718647, 44084638, 44084638],
        3: ['AAK1', 'AAED1', 'AAAS', 'XRCC1', 'foo'],
        4: [True, True, True, 1, 1],
        5: [True, True, False, 0, 1]})
    with patch.object(genie_registry.bed, "create_gtf",
                      return_value=(EXON_TEMP.name, GENE_TEMP.name)):
        new_beddf = bed_class._process(
            beddf, seq_assay_id, new_path, parentid, create_panel=False)
        new_beddf.sort_values("ID", inplace=True)
        new_beddf.reset_index(drop=True, inplace=True)
        assert_frame_equal(expected_beddf,
                           new_beddf[expected_beddf.columns],
                           check_dtype=False)


def test_includeinpanel___process():
    """
    Make sure includeInPanel column is propogated and
    intergenic region is captured
    """

    expected_beddf = pd.DataFrame(dict(
        Chromosome=['2', '9', '12', '19'],
        Start_Position=[69688432, 1111, 53700240, 44080953],
        End_Position=[69689532, 1111, 53719548, 44084624],
        Hugo_Symbol=['AAK1', float('nan'), 'AAAS', float('nan')],
        includeInPanel=[True, True, False, True],
        clinicalReported=[float('nan')]*4,
        ID=['foo', 'bar', 'baz', 'boo'],
        SEQ_ASSAY_ID=['SAGE-TEST', 'SAGE-TEST', 'SAGE-TEST', 'SAGE-TEST'],
        Feature_Type=['exon', 'intergenic', 'exon', 'exon'],
        CENTER=['SAGE', 'SAGE', 'SAGE', 'SAGE']))

    expected_beddf.sort_values("Chromosome", inplace=True)
    expected_beddf.reset_index(drop=True, inplace=True)

    # symbols that can't be map should be null,
    # includeInPanel column should be included if it exists
    beddf = pd.DataFrame({
        0: ['2', '9', '12', '19'],
        1: [69688432, 1111, 53700240, 44080953],
        2: [69689532, 1111, 53719548, 44084624],
        3: ['foo', 'bar', 'baz', 'boo'],
        4: [True, True, 0, 1]})
    with patch.object(genie_registry.bed, "create_gtf",
                      return_value=(EXON_TEMP.name, GENE_TEMP.name)):
        new_beddf = bedsp_class._process(
            beddf, seq_assay_id, new_path, parentid, create_panel=False)
        new_beddf.sort_values("Chromosome", inplace=True)
        new_beddf.reset_index(drop=True, inplace=True)
        assert_frame_equal(expected_beddf,
                        new_beddf[expected_beddf.columns],
                        check_dtype=False)


def test_clinicalreport___process():
    expected_beddf = pd.DataFrame(dict(
        Chromosome=['2', '9', '12', '19'],
        Start_Position=[69688432, 1111, 53700240, 44080953],
        End_Position=[69689532, 1111, 53719548, 44084624],
        Hugo_Symbol=['AAK1', float('nan'), 'AAAS', float('nan')],
        includeInPanel=[True, True, False, True],
        clinicalReported=[float('nan')]*4,
        ID=['foo', 'bar', 'baz', 'boo'],
        SEQ_ASSAY_ID=['SAGE-TEST', 'SAGE-TEST', 'SAGE-TEST', 'SAGE-TEST'],
        Feature_Type=['exon', 'intergenic', 'exon', 'exon'],
        CENTER=['SAGE', 'SAGE', 'SAGE', 'SAGE']))

    expected_beddf.sort_values("Chromosome", inplace=True)
    expected_beddf.reset_index(drop=True, inplace=True)

    # symbols that can't be map should be null,
    # includeInPanel column should be included if it exists
    beddf = pd.DataFrame({
        0: ['2', '9', '12', '19'],
        1: [69688432, 1111, 53700240, 44080953],
        2: [69689532, 1111, 53719548, 44084624],
        3: ['foo', 'bar', 'baz', 'boo'],
        4: [True, True, False, True],
        5: [True, float('nan'), False, True]})
    with patch.object(genie_registry.bed, "create_gtf",
                      return_value=(EXON_TEMP.name, GENE_TEMP.name)):
        new_beddf = bedsp_class._process(
            beddf, seq_assay_id, new_path, parentid, create_panel=False)
        new_beddf.sort_values("Chromosome", inplace=True)
        new_beddf.reset_index(drop=True, inplace=True)
        assert_frame_equal(expected_beddf,
                        new_beddf[expected_beddf.columns],
                        check_dtype=False)


def test_filetype():
    assert bed_class._fileType == "bed"
    assert bedsp_class._fileType == "bedSP"


@pytest.fixture(params=[
    (["foo"]),
    (["SAGE-test.txt"]),
    (["foo"]),
    (["nonGENIE_SAGE-test.txt"])
    ])
def filename_fileformat_map(request):
    return request.param


def test_incorrect_validatefilename(filename_fileformat_map):
    filepath_list = filename_fileformat_map
    with pytest.raises(AssertionError):
        bed_class.validateFilename(filepath_list)


def test_correct_validatefilename():
    assert bed_class.validateFilename(["SAGE-test.bed"]) == "bed"
    assert bedsp_class.validateFilename(["nonGENIE_SAGE-test.bed"]) == "bedSP"


def test_perfect__validate():
    bedDf = pd.DataFrame(dict(
        a=['2', '9', '12'],
        b=[69688533, 99401860, 53701241],
        c=[69901480, 99417584, 53718647],
        d=['AAK1', 'AAED1', 'AAAS'],
        e=[True, True, False],
        f=[True, True, False]))

    error, warning = bed_class._validate(bedDf)
    assert error == ""
    assert warning == ""


def test_90percent_boundary__validate():
    bedDf = pd.DataFrame(dict(
        a=['2', '9', '12'],
        b=[69688432, 99416585, 53700240],
        c=[69689532, 99417685, 53719548],
        d=['AAK1', 'AAED1', 'AAAS'],
        e=[True, True, False],
        f=[True, True, False]))
    error, warning = bed_class._validate(bedDf)
    assert error == ""
    assert warning == ""


def test_missingcols_failure__validate():
    emptydf = pd.DataFrame()
    error, warning = bed_class._validate(emptydf)
    expected_errors = (
        "BED file: Must at least have five columns in this order: "
        "Chromosome, Start_Position, End_Position, Hugo_Symbol, "
        "includeInPanel. Make sure there are no headers.\n")
    assert error == expected_errors
    assert warning == ""


def test_hugosymbol_failure__validate():
    bedDf = pd.DataFrame(dict(
        a=['2', '9', '12'],
        b=[69688533, 99401860, 53701241],
        c=[69901480, 99417584, 53718647],
        d=['+', float('nan'), 'AAAS'],
        e=[True, True, False]))
    error, warning = bed_class._validate(bedDf)
    expected_errors = (
        "BED file: You cannot submit any null symbols.\n"
        "BED file: Fourth column must be the Hugo_Symbol column, "
        "not the strand column\n")
    assert error == expected_errors
    assert warning == ""


def test_badinputs_failure__validate():
    bedDf = pd.DataFrame(dict(
        a=['2', '9', '12'],
        b=['69688533', 99401860, 53701241],
        c=[69901480, '99417584', 53718647],
        d=['AAK1', 'AAED1', 'AAAS'],
        e=[True, False, "foobar"]))

    error, warning = bed_class._validate(bedDf)
    expected_errors = (
        "BED file: The Start_Position column must only be integers. "
        "Make sure there are no headers.\n"
        "BED file: The End_Position column must only be integers. "
        "Make sure there are no headers.\n"
        "BED file: Please double check your includeInPanel column.  "
        "This column must only be these values: True, False\n")
    assert error == expected_errors
    assert warning == ""


def test_90percentboundary_failure__validate():
    # Test 90% boundary failure boundary, with incorrect gene names
    bedDf = pd.DataFrame(dict(
        a=['2', '9', '12'],
        b=[69901381, 4345, 11111],
        c=[69911481, 99417590, 11113],
        d=['foo', 'foo', 'AAAS'],
        e=[True, True, False]))

    error, warning = bedsp_class._validate(bedDf)
    expected_errors = (
        "BED file: You have no correct gene symbols. "
        "Make sure your gene symbol column (4th column) is formatted like so: "
        "SYMBOL(;optionaltext).  Optional text can be semi-colon separated.\n")
    expected_warnings = (
        "BED file: Any gene names that can't be remapped will be null.\n")
    assert error == expected_errors
    assert warning == expected_warnings


def test_overlapping__validate():
    """
    Check if genes that overlap have no errors
    - Submitted bed start position in a gene
    - Submitted bed end position in a gene
    - Submitted bed region surrounds a gene
    """
    beddf = pd.DataFrame(dict(
        a=['2', '9', '2'],
        b=[1111, 4345, 69901281],
        c=[69880186, 99417590, 70000000],
        d=['AAK1', 'AAED1', 'AAK1'],
        e=[True, False, True]))
    error, warning = bedsp_class._validate(beddf)
    assert error == ""
    assert warning == ""


def test_symbolnull_failure__validate():
    # Test 2 gene symbols returned NULL
    bedDf = pd.DataFrame(dict(
        a=['19'],
        b=[44080953],
        c=[44084624],
        d=['AAK1'],
        e=[False]))

    error, warning = bedsp_class._validate(bedDf)
    expected_errors = (
        "BED file: You have no correct gene symbols. "
        "Make sure your gene symbol column (4th column) is formatted like so: "
        "SYMBOL(;optionaltext).  Optional text can be semi-colon separated.\n")
    expected_warnings = (
        "BED file: Any gene names that can't be remapped will be null.\n")
    assert error == expected_errors
    assert warning == expected_warnings



"""Updates dashboard tables"""
import argparse
import datetime
import logging
import os

import pandas as pd
import synapseclient
from synapseclient.core.utils import to_unix_epoch_time

from genie import process_functions

logger = logging.getLogger(__name__)


def get_center_data_completion(center, df):
    """
    Gets center data completion.  Calulates the percentile of
    how complete a clinical data element is:
    Number of not blank/Unknown/NA divded by
    total number of patients or samples

    Args:
        center: GENIE center
        df: sample or patient dataframe

    Returns:
        Dataframe: Center data
    """
    centerdf = df[df['CENTER'] == center]
    total = len(centerdf)
    center_data = pd.DataFrame()
    skip_cols = ['CENTER', 'PATIENT_ID', 'SAMPLE_ID', 'SAMPLE_TYPE_DETAILED',
                 'SECONDARY_RACE', 'TERTIARY_RACE']
    for col in centerdf:
        if col not in skip_cols:
            not_missing = [not pd.isnull(value) and value != 'Not Collected'
                           for value in centerdf[col]]
            completeness = float(sum(not_missing)) / int(total)
            returned = pd.DataFrame([[col, center, total, completeness]])
            center_data = center_data.append(returned)
    return center_data


def update_samples_in_release_table(syn, file_mapping, release,
                                    samples_in_release_synid):
    '''
    Updates the sample in release table
    This tracks the samples of each release.  1 means it exists, and 0
    means it doesn't

    Args:
        syn: synapse object
        file_mapping: file mapping generated from file mapping function
        release:  GENIE release number (ie. 5.3-consortium)
        samples_in_release_synid: Synapse Id of 'samples in release' Table
    '''
    clinical_ent = syn.get(file_mapping['clinical'], followLink=True)
    clinicaldf = pd.read_csv(clinical_ent.path, sep="\t", comment="#")
    cols = [i['name'] for i in
            list(syn.getTableColumns(samples_in_release_synid))]

    if release not in cols:
        schema = syn.get(samples_in_release_synid)
        syn_col = synapseclient.Column(name=release, columnType='INTEGER',
                                       defaultValue=0)
        new_column = syn.store(syn_col)
        schema.addColumn(new_column)
        schema = syn.store(schema)
    # Columns of samples in release
    samples_per_release = syn.tableQuery(
        'SELECT SAMPLE_ID, "{}" FROM {}'.format(release,
                                                samples_in_release_synid))

    samples_per_releasedf = samples_per_release.asDataFrame()
    new_samples = clinicaldf[['SAMPLE_ID']][
        ~clinicaldf.SAMPLE_ID.isin(samples_per_releasedf.SAMPLE_ID)]

    new_samples[release] = 1
    old_samples = clinicaldf[['SAMPLE_ID']][
        clinicaldf.SAMPLE_ID.isin(samples_per_releasedf.SAMPLE_ID)]

    old_samples[release] = 1
    samples_in_releasedf = new_samples.append(old_samples)
    process_functions.updateDatabase(syn, samples_per_releasedf,
                                     samples_in_releasedf,
                                     samples_in_release_synid, ["SAMPLE_ID"])


def update_cumulative_sample_table(syn, file_mapping, release,
                                   cumulative_sample_count_synid):
    '''
    Consortium release sample count table update function
    This gets the cumulative sample count of each file type in each release

    Args:
        syn: synapse object
        file_mapping: file mapping generated from file mapping function
        release:  GENIE release number (ie. 5.3-consortium)
        cumulative_sample_count_synid: Synapse Id of
            'Cumulative sample count' Table
    '''

    sample_count_per_round = syn.tableQuery(
        "SELECT * FROM {} where Release = '{}'".format(cumulative_sample_count_synid,
                                                       release))
    sample_count_per_rounddf = sample_count_per_round.asDataFrame()

    clinical_ent = syn.get(file_mapping['clinical'], followLink=True)
    clinicaldf = pd.read_csv(clinical_ent.path, sep="\t", comment="#")
    clinicaldf.columns = [i.upper() for i in clinicaldf.columns]
    if clinicaldf.get("CENTER") is None:
        clinicaldf['CENTER'] = \
            [sample.split("-")[1] for sample in clinicaldf.SAMPLE_ID]
    clinical_counts = clinicaldf['CENTER'].value_counts()
    clinical_counts['Total'] = sum(clinical_counts)
    clinical_counts.name = "Clinical"

    fusion_ent = syn.get(file_mapping['fusion'], followLink=True)
    fusiondf = pd.read_csv(fusion_ent.path, sep="\t", comment="#")
    fusiondf.columns = [i.upper() for i in fusiondf.columns]
    fusion_counts = fusiondf['CENTER'][
        ~fusiondf['TUMOR_SAMPLE_BARCODE'].duplicated()].value_counts()
    fusion_counts['Total'] = sum(fusion_counts)

    cna_ent = syn.get(file_mapping['cna'], followLink=True)
    cnadf = pd.read_csv(cna_ent.path, sep="\t", comment="#")
    cna_counts = pd.Series(
        [i.split("-")[1] for i in cnadf.columns[1:]]).value_counts()
    cna_counts['Total'] = sum(cna_counts)

    seg_ent = syn.get(file_mapping['seg'], followLink=True)
    segdf = pd.read_csv(seg_ent.path, sep="\t", comment="#")
    segdf.columns = [i.upper() for i in segdf.columns]

    segdf['CENTER'] = [i.split("-")[1] for i in segdf['ID']]
    seg_counts = segdf['CENTER'][~segdf['ID'].duplicated()].value_counts()
    seg_counts['Total'] = sum(seg_counts)

    total_counts = pd.DataFrame(clinical_counts)
    total_counts['Fusions'] = fusion_counts
    total_counts['CNV'] = cna_counts
    total_counts['Mutation'] = clinical_counts
    total_counts['SEG'] = seg_counts
    total_counts = total_counts.fillna(0)
    total_counts = total_counts.applymap(int)
    total_counts['Center'] = total_counts.index
    total_counts['Release'] = release
    process_functions.updateDatabase(syn, sample_count_per_rounddf,
                                     total_counts,
                                     cumulative_sample_count_synid,
                                     ["Center", "Release"],
                                     to_delete=True)


def get_file_mapping(syn, release_folder_synid):
    """
    Get file mapping between important files needed for dashboard and
    their synapse ids

    Args:
        syn:  synapse object
        release_folder_synid: synapse id of release

    """
    files = syn.getChildren(release_folder_synid)
    file_mapping = dict()
    for metadata in files:
        filename = metadata['name']
        synid = metadata['id']
        if not filename.startswith("meta"):
            if filename.startswith("data_clinical_sample"):
                file_mapping['clinical'] = synid
            elif filename.endswith("fusions.txt"):
                file_mapping['fusion'] = synid
            elif filename.endswith("CNA.txt"):
                file_mapping['cna'] = synid
            elif filename.endswith(".seg"):
                file_mapping['seg'] = synid
    return file_mapping


def update_release_numbers(syn, database_mappingdf, release=None):
    """
    Updates all release dashboard numbers or specific release number

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
        release: GENIE release (ie. 5.3-consortium).  Defaults to None
    """
    # Update release table with current release or all releases
    samples_in_release_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'samplesInRelease'].values[0]
    cumulative_sample_count_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'cumulativeSampleCount'].values[0]

    release_folder_fileview_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'releaseFolder'].values[0]
    release_folder = syn.tableQuery(
        "select id,name from %s" % release_folder_fileview_synid +
        " where name not like 'Release%' and name <> 'case_lists' and " +
        "name not like '%.0.%'")
    release_folderdf = release_folder.asDataFrame()

    for rel_synid, rel_name in zip(release_folderdf.id, release_folderdf.name):
        file_mapping = get_file_mapping(syn, rel_synid)
        # If release is specified, only process on that,
        # otherwise process for all
        if release is None or release == rel_name:
            update_samples_in_release_table(
                syn, file_mapping, rel_name, samples_in_release_synid)
            update_cumulative_sample_table(
                syn, file_mapping, rel_name, cumulative_sample_count_synid)
        else:
            pass


def update_database_numbers(syn, database_mappingdf):
    """
    Updates database cumulative numbers (Only called when not staging)

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    cumulative_sample_count_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'cumulativeSampleCount'].values[0]
    # Database
    database_count = syn.tableQuery(
        "SELECT * FROM %s where Release = 'Database'" %
        cumulative_sample_count_synid)
    database_countdf = database_count.asDataFrame()
    clinical = syn.tableQuery('select CENTER from syn7517674')
    clinicaldf = clinical.asDataFrame()
    clinincal_counts = clinicaldf['CENTER'].value_counts()
    clinincal_counts['Total'] = sum(clinincal_counts)
    clinincal_counts.name = "Clinical"

    fusion = syn.tableQuery('select * from syn7893268')
    fusiondf = fusion.asDataFrame()
    fusion_counts = fusiondf['CENTER'][
        ~fusiondf['TUMOR_SAMPLE_BARCODE'].duplicated()].value_counts()
    fusion_counts['Total'] = sum(fusion_counts)

    center_flat_files = syn.getChildren("syn12278118")
    cna_file_paths = [syn.get(file['id']).path for file in center_flat_files if
                      file['name'].startswith("data_CNA")]
    cna_numbers = {}
    for cna_file in cna_file_paths:
        center = os.path.basename(cna_file).replace(".txt", "").split("_")[2]
        with open(cna_file, 'r') as cna:
            header = cna.readline()
            samples = header.split("\t")
            # Minus one because of Hugo_Symbol
            cna_numbers[center] = len(samples) - 1
    cna_counts = pd.Series(cna_numbers)
    cna_counts['Total'] = sum(cna_counts)

    seg = syn.tableQuery('select * from syn7893341')
    segdf = seg.asDataFrame()
    seg_counts = segdf['CENTER'][~segdf['ID'].duplicated()].value_counts()
    seg_counts['Total'] = sum(seg_counts)

    db_counts = pd.DataFrame(clinincal_counts)
    db_counts['Fusions'] = fusion_counts
    db_counts['CNV'] = cna_counts
    db_counts['Mutation'] = clinincal_counts
    db_counts['SEG'] = seg_counts
    db_counts = db_counts.fillna(0)
    db_counts = db_counts.applymap(int)
    db_counts['Center'] = db_counts.index
    db_counts['Release'] = "Database"
    process_functions.updateDatabase(syn, database_countdf, db_counts,
                                     cumulative_sample_count_synid,
                                     ["Center", "Release"])
    today = datetime.date.today()
    if today.month in [1, 4, 8, 12]:
        db_count_tracker = db_counts[['Clinical', 'Center', 'Release']]
        db_count_tracker.rename(
            columns={'Clinical': 'sample_count',
                     'Center': 'center',
                     'Release': 'date'},
            inplace=True)
        db_count_tracker['date'] = today.strftime("%b-%Y")
        # Hard coded syn id
        syn.store(synapseclient.Table("syn18404852", db_count_tracker))


def update_oncotree_code_tables(syn, database_mappingdf):
    """
    Updates database statistics of oncotree codes
    and primary onocotree codes

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    oncotree_distribution_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'oncotree'].values[0]

    clinical = syn.tableQuery('select * from syn7517674')
    clinicaldf = clinical.asDataFrame()

    # DISTRIBUTION OF ONCOTREE CODE TABLE UPDATE
    oncotree_code_distributiondf = pd.DataFrame(
        columns=set(clinicaldf['CENTER']),
        index=set(clinicaldf['ONCOTREE_CODE']))
    for center in oncotree_code_distributiondf.columns:
        onc_counts = clinicaldf['ONCOTREE_CODE'][
            clinicaldf['CENTER'] == center].value_counts()
        oncotree_code_distributiondf[center] = onc_counts
    oncotree_code_distributiondf = oncotree_code_distributiondf.fillna(0)
    oncotree_code_distributiondf = oncotree_code_distributiondf.applymap(int)
    oncotree_code_distributiondf['Total'] = \
        oncotree_code_distributiondf.apply(sum, axis=1)
    oncotree_code_distributiondf['Oncotree_Code'] = \
        oncotree_code_distributiondf.index

    oncotree_distribution_db = syn.tableQuery(
        'SELECT %s FROM %s' %
        ("Oncotree_Code," + ",".join(clinicaldf['CENTER'].unique()) +
         ",Total", oncotree_distribution_synid))

    oncotree_distribution_dbdf = oncotree_distribution_db.asDataFrame()
    process_functions.updateDatabase(syn, oncotree_distribution_dbdf,
                                     oncotree_code_distributiondf,
                                     oncotree_distribution_synid,
                                     ["Oncotree_Code"], to_delete=True)

    # DISTRIBUTION OF PRIMARY CODE TABLE UPDATE
    oncotree_link_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'oncotreeLink'].values[0]
    primary_code_synId = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'primaryCode'].values[0]

    # Can also use most up to date oncotree code,
    # because these tables are updated from the database
    oncotree_link_ent = syn.get(oncotree_link_synid)
    oncotree_link = oncotree_link_ent.externalURL
    oncotree_mapping = \
        process_functions.get_oncotree_code_mappings(oncotree_link)

    clinicaldf['PRIMARY_CODES'] = \
        [oncotree_mapping[i.upper()]['ONCOTREE_PRIMARY_NODE']
         if i.upper() in oncotree_mapping.keys() else 'DEPRECATED_CODE'
         for i in clinicaldf.ONCOTREE_CODE]

    # ### DISTRIBUTION OF PRIMARY ONCOTREE CODE TABLE UPDATE
    primary_code_distributiondf = pd.DataFrame(
        columns=set(clinicaldf['CENTER']),
        index=set(clinicaldf['PRIMARY_CODES']))

    for center in primary_code_distributiondf.columns:
        onc_counts = clinicaldf['PRIMARY_CODES'][
            clinicaldf['CENTER'] == center].value_counts()
        primary_code_distributiondf[center] = onc_counts
    primary_code_distributiondf = primary_code_distributiondf.fillna(0)
    primary_code_distributiondf = primary_code_distributiondf.applymap(int)
    primary_code_distributiondf['Total'] = \
        primary_code_distributiondf.apply(sum, axis=1)
    primary_code_distributiondf['Oncotree_Code'] = \
        primary_code_distributiondf.index

    primary_code_dist_db = syn.tableQuery(
        'SELECT %s FROM %s' %
        ("Oncotree_Code," + ",".join(clinicaldf['CENTER'].unique()) +
         ",Total", primary_code_synId))

    primary_code_dist_dbdf = primary_code_dist_db.asDataFrame()
    process_functions.updateDatabase(syn, primary_code_dist_dbdf,
                                     primary_code_distributiondf,
                                     primary_code_synId, ["Oncotree_Code"],
                                     to_delete=True)


def update_sample_difference_table(syn, database_mappingdf):
    """
    Updates sample difference table between
    consortium releases

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    cumulative_sample_count_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'cumulativeSampleCount'].values[0]

    sample_diff_count_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'sampleDiffCount'].values[0]

    # UPDATE DIFF TABLE
    sample_count_per_round = syn.tableQuery(
        "SELECT * FROM %s where Center <> 'Total' and Release <> 'Database'"
        % cumulative_sample_count_synid)

    sample_count_per_rounddf = sample_count_per_round.asDataFrame()
    releases = list(sample_count_per_rounddf['Release'].unique())
    # sort the releases and remove public releases
    releases.sort()
    consortium_releases = [
        release for release in releases if "public" not in release
        and ".0." not in release]

    diff_between_releasesdf = sample_count_per_rounddf[
        sample_count_per_rounddf['Release'] == consortium_releases[0]]

    for index, release_name in enumerate(consortium_releases[1:]):
        prior_release = sample_count_per_rounddf[
            sample_count_per_rounddf['Release'] == consortium_releases[index]]

        current_release = sample_count_per_rounddf[
            sample_count_per_rounddf['Release'] == release_name]

        prior_release.index = prior_release['Center']
        current_release.index = current_release['Center']

        del prior_release['Center']
        del prior_release['Release']
        del current_release['Center']
        del current_release['Release']
        # Append new rows of centers that are new and
        # just added to the releases
        new_centers = current_release.index[
            ~current_release.index.isin(prior_release.index)]

        if not new_centers.empty:
            prior_release = prior_release.append(
                pd.DataFrame(index=new_centers))
            prior_release = prior_release.fillna(0)
        difference = current_release - prior_release
        difference['Center'] = difference.index
        difference['Release'] = release_name
        diff_between_releasesdf = diff_between_releasesdf.append(difference)

    difftable_db = syn.tableQuery('SELECT * FROM %s' % sample_diff_count_synid)
    difftable_dbdf = difftable_db.asDataFrame()
    difftable_dbdf = difftable_dbdf.fillna(0)
    new_values = diff_between_releasesdf[[
        'Clinical', 'Mutation',
        'CNV', 'SEG', 'Fusions']].fillna(0).applymap(int)

    diff_between_releasesdf[
        ['Clinical', 'Mutation', 'CNV', 'SEG', 'Fusions']] = new_values

    process_functions.updateDatabase(syn, difftable_dbdf,
                                     diff_between_releasesdf,
                                     sample_diff_count_synid,
                                     ["Center", "Release"],
                                     to_delete=True)


def update_data_completeness_table(syn, database_mappingdf):
    """
    Updates the data completeness of the database

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    data_completion_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'dataCompletion'].values[0]

    sample = syn.tableQuery('select * from syn7517674')
    sampledf = sample.asDataFrame()
    patient = syn.tableQuery('select * from syn7517669')
    patientdf = patient.asDataFrame()

    data_completenessdf = pd.DataFrame()
    center_infos = sampledf.CENTER.drop_duplicates().apply(
        lambda center: get_center_data_completion(center, sampledf))
    for center_info in center_infos:
        data_completenessdf = data_completenessdf.append(center_info)

    center_infos = patientdf.CENTER.drop_duplicates().apply(
        lambda center: get_center_data_completion(center, patientdf))
    for center_info in center_infos:
        data_completenessdf = data_completenessdf.append(center_info)

    data_completeness_db = syn.tableQuery(
        'select * from %s' % data_completion_synid)
    data_completeness_dbdf = data_completeness_db.asDataFrame()
    data_completenessdf.columns = data_completeness_dbdf.columns
    process_functions.updateDatabase(syn, data_completeness_dbdf,
                                     data_completenessdf,
                                     data_completion_synid,
                                     ["FIELD", "CENTER"],
                                     to_delete=True)


def update_wiki(syn, database_mappingdf):
    """
    Updates the GENIE project dashboard wiki timestamp

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database

    """
    # Updates to query and date dashboard was updated
    cumulative_sample_count_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'cumulativeSampleCount'].values[0]

    primary_code_synId = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'primaryCode'].values[0]

    centers = syn.tableQuery(
        'select distinct(CENTER) as CENTER from syn7517674')

    centersdf = centers.asDataFrame()
    now = datetime.datetime.now()
    markdown = \
        ["_Updated {month}/{day}/{year}_\n\n".format(
            month=now.month,
            day=now.day,
            year=now.year),
         "##Count of Clinical Samples\n",
         "${synapsetable?query=SELECT Center%2C Clinical%2C Release FROM " +
            cumulative_sample_count_synid + "}\n\n",
         "\n\n##Primary Oncotree Codes\n\n",
         "${synapsetable?query=SELECT Oncotree%5FCode%2C " +
            "%2C ".join(centersdf['CENTER'].unique()) +
            "%2C Total FROM " + primary_code_synId +
            " ORDER BY Total DESC&limit=15}\n\n"]

    wikipage = syn.getWiki("syn3380222", 235803)
    wikipage.markdown = "".join(markdown)
    syn.store(wikipage)


def string_to_unix_epoch_time_milliseconds(string_time):
    """
    Takes dates in this format: 2018-10-25T20:16:07.959Z
    and turns it into unix epoch time in milliseconds

    Args:
        string_time: string in this format: 2018-10-25T20:16:07.959Z

    Returns:
        unix epoch time in milliseconds
    """
    datetime_obj = datetime.datetime.strptime(
        string_time.split(".")[0], "%Y-%m-%dT%H:%M:%S")
    return to_unix_epoch_time(datetime_obj)


def update_data_release_file_table(syn, database_mappingdf):
    """
    Updates data release file table

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    release_folder_fileview_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'releaseFolder'].values[0]
    release_folder = syn.tableQuery(
        "select id,name from %s" % release_folder_fileview_synid +
        " where name not like 'Release%' and name <> 'case_lists' " +
        "and name not like '0.%'")
    release_folderdf = release_folder.asDataFrame()

    data_release_table_synid = "syn16804261"
    data_release_table = syn.tableQuery(
        "select * from %s" % data_release_table_synid)
    data_release_tabledf = data_release_table.asDataFrame()

    not_in_release_tabledf = release_folderdf[
        ~release_folderdf.name.isin(data_release_tabledf.release)]

    for synid, name in zip(not_in_release_tabledf.id,
                           not_in_release_tabledf.name):
        release_files = syn.getChildren(synid)

        append_rows = [
            [release_file['name'],
             release_file['id'],
             name,
             string_to_unix_epoch_time_milliseconds(
                release_file['modifiedOn']), synid]
            for release_file in release_files
            if release_file['name'] != "case_lists"]

        syn.store(synapseclient.Table(data_release_table_synid, append_rows))


def check_column_decreases(currentdf, olderdf):
    """
    Checks entity decreases

    Args:
        current_ent: Current entity dataframe
        old_ent: Older entity dataframe

    Returns:
        Differences in values
    """
    diff_map = dict()
    for col in currentdf:
        new_counts = currentdf[col].value_counts()
        if olderdf.get(col) is not None:
            old_counts = olderdf[col].value_counts()
            # Make sure any values that exist in the new get added
            # to the old to show the decrease
            new_keys = pd.Series(index=new_counts.keys()[
                ~new_counts.keys().isin(old_counts.keys())])
            old_counts = old_counts.add(new_keys, fill_value=0)
            old_counts.fillna(0, inplace=True)
            # Make sure any values that don't exist in the old get added
            # to show the decrease
            new_keys = pd.Series(index=old_counts.keys()[
                ~old_counts.keys().isin(new_counts.keys())])
            new_counts = new_counts.add(new_keys, fill_value=0)
            new_counts.fillna(0, inplace=True)
            if any(new_counts - old_counts < 0):
                logger.info("\tDECREASE IN COLUMN: %s" % col)
                # diff = new_counts[new_counts - old_counts < 0]
                diffs = new_counts-old_counts
                diffstext = diffs[diffs < 0].to_csv().replace("\n", "; ")
                logger.info("\t" + diffstext)
                diff_map[col] = True
            else:
                diff_map[col] = False
    return diff_map


def print_clinical_values_difference_table(syn, database_mappingdf):
    """
    Checks for a decrease in values in the clinical file
    from last consortium release to most recent consortium release

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    release_folder_fileview_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'releaseFolder'].values[0]

    clinical_key_decrease_synid = database_mappingdf['Id'][
        database_mappingdf['Database'] == 'clinicalKeyDecrease'].values[0]

    release_folder = syn.tableQuery(
        "select id,name from %s" % release_folder_fileview_synid +
        " where name not like 'Release%' and name <> 'case_lists' " +
        "and name not like '%.0.%' and name not like '%-public'")

    release_folderdf = release_folder.asDataFrame()
    release_folderdf.sort_values("name", ascending=False, inplace=True)
    current_release = release_folderdf['id'][0]
    older_release = release_folderdf['id'][1]

    current_release_files = syn.getChildren(current_release)
    current_clinical_synids = {
        file['name']: file['id']
        for file in current_release_files if file['name'] in
        ['data_clinical_sample.txt', 'data_clinical_patient.txt']}

    older_release_files = syn.getChildren(older_release)

    older_clinical_synids = {
        file['name']: file['id']
        for file in older_release_files if file['name'] in
        ['data_clinical_sample.txt', 'data_clinical_patient.txt']}

    current_sample_ent = syn.get(
        current_clinical_synids['data_clinical_sample.txt'], followLink=True)

    older_sample_ent = syn.get(
        older_clinical_synids['data_clinical_sample.txt'], followLink=True)
    current_sampledf = pd.read_csv(
        current_sample_ent.path, sep="\t", comment="#")

    current_sampledf['CENTER'] = [
        patient.split("-")[1] for patient in current_sampledf['PATIENT_ID']]

    older_sampledf = pd.read_csv(older_sample_ent.path, sep="\t", comment="#")
    older_sampledf['CENTER'] = [
        patient.split("-")[1] for patient in older_sampledf['PATIENT_ID']]
    # Rather than take the CENTER, must take the SAMPLE_ID to compare
    current_sampledf = current_sampledf[current_sampledf['SAMPLE_ID'].isin(
        older_sampledf['SAMPLE_ID'].unique())]

    logger.info("SAMPLE CLINICAL VALUE DECREASES")
    center_decrease_mapping = dict()
    for center in older_sampledf['CENTER'].unique():
        current_center_sampledf = current_sampledf[
            current_sampledf['CENTER'] == center]

        older_center_sampledf = older_sampledf[
            older_sampledf['CENTER'] == center]

        logger.info(center)

        decrease_map = check_column_decreases(
            current_center_sampledf, older_center_sampledf)
        center_decrease_mapping[center] = decrease_map

    current_patient_ent = syn.get(
        current_clinical_synids['data_clinical_patient.txt'], followLink=True)

    older_patient_ent = syn.get(
        older_clinical_synids['data_clinical_patient.txt'], followLink=True)

    current_patientdf = pd.read_csv(
        current_patient_ent.path, sep="\t", comment="#")

    older_patientdf = pd.read_csv(
        older_patient_ent.path, sep="\t", comment="#")
    # Rather than take the CENTER, must take the PATIENT_ID to compare
    current_patientdf = current_patientdf[current_patientdf['PATIENT_ID'].isin(
        older_patientdf['PATIENT_ID'].unique())]

    logger.info("PATIENT CLINICAL VALUE DECREASES")
    for center in older_patientdf['CENTER'].unique():
        current_center_patientdf = current_patientdf[
            current_patientdf['CENTER'] == center]

        older_center_patientdf = older_patientdf[
            older_patientdf['CENTER'] == center]

        logger.info(center)
        patient_decrease_map = check_column_decreases(
            current_center_patientdf, older_center_patientdf)

        center_decrease_mapping[center].update(patient_decrease_map)

    center_decrease_mapping = pd.DataFrame(center_decrease_mapping)
    center_decrease_mapping = center_decrease_mapping.transpose()
    center_decrease_mapping['CENTER'] = center_decrease_mapping.index

    clinical_key_decrease = syn.tableQuery("select * from {0}".format(
        clinical_key_decrease_synid))
    clinical_key_decreasedbdf = clinical_key_decrease.asDataFrame()
    process_functions.updateDatabase(syn, clinical_key_decreasedbdf,
                                     center_decrease_mapping,
                                     clinical_key_decrease_synid, ["CENTER"],
                                     to_delete=True)


def run_dashboard(syn, database_mappingdf, release, staging=False,
                  public=False):
    """
    Runs the dashboard scripts

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
        release: GENIE release (ie. 5.3-consortium)

    """
    update_release_numbers(syn, database_mappingdf, release=release)

    if not staging:
        update_data_release_file_table(syn, database_mappingdf)
        if not public:
            print_clinical_values_difference_table(syn, database_mappingdf)
            update_sample_difference_table(syn, database_mappingdf)
            update_data_completeness_table(syn, database_mappingdf)
            update_database_numbers(syn, database_mappingdf)
            update_oncotree_code_tables(syn, database_mappingdf)
            update_wiki(syn, database_mappingdf)


def main():
    parser = argparse.ArgumentParser(description='Update dashboard tables')

    parser.add_argument(
        '--release',
        help="GENIE release number (ie. 5.3-consortium)",
        default=None)

    parser.add_argument(
        "--pem_file",
        type=str,
        help="Path to PEM file (genie.pem)")

    parser.add_argument(
        "--staging",
        action='store_true',
        help="Using staging directory files")

    parser.add_argument(
        "--debug",
        action='store_true',
        help="Synapse debugging flag")

    parser.add_argument(
        "--public",
        action='store_true',
        help="Set true if releasing public release")

    args = parser.parse_args()
    syn = process_functions.synLogin(args)
    if args.staging:
        # Database to Synapse Id mapping Table
        database_mapping_synid = 'syn12094210'
    else:
        database_mapping_synid = 'syn10967259'

    database_mapping = syn.tableQuery(
        'select * from %s' % database_mapping_synid)
    database_mappingdf = database_mapping.asDataFrame()

    run_dashboard(syn, database_mappingdf, args.release,
                  staging=args.staging, public=args.public)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
"""Script to crawl Synapse folder for a center, validate, and update database tables.

"""
import argparse
from datetime import date
import logging
import os

from genie import (input_to_database, write_invalid_reasons,
                   process_functions, config)

logger = logging.getLogger(__name__)

# TODO: Remove oncotree_link
# TODO: Remove gneie_annotation_pkg
def main(process,
         project_id,
         center=None,
         pemfile=None,
         delete_old=False,
         only_validate=False,
         oncotree_link=None,
         genie_annotation_pkg=None,
         create_new_maf_database=False,
         debug=False,
         format_registry=None):

    syn = process_functions.synLogin(pemfile, debug=debug)

    # Get the Synapse Project where data is stored
    # Should have annotations to find the table lookup
    project = syn.get(project_id)
    database_to_synid_mapping_synid = project.annotations.get("dbMapping", "")
    
    databaseToSynIdMapping = syn.tableQuery(
        'SELECT * FROM {}'.format(database_to_synid_mapping_synid[0]))
    databaseToSynIdMappingDf = databaseToSynIdMapping.asDataFrame()

    center_mapping_id = process_functions.getDatabaseSynId(
        syn, "centerMapping",
        databaseToSynIdMappingDf=databaseToSynIdMappingDf)

    center_mapping = syn.tableQuery('SELECT * FROM %s' % center_mapping_id)
    center_mapping_df = center_mapping.asDataFrame()

    if center is not None:
        assert center in center_mapping_df.center.tolist(), (
            "Must specify one of these centers: {}".format(
                ", ".join(center_mapping_df.center)))
        centers = [center]
    else:
        # exclude_sites = ['JHU', 'DFCI', 'GRCC', 'VICC', 'NKI', 'MSK',
        #                  'UHN', 'MDA', 'WAKE', 'YALE', 'UCSF', 'CRUK',
        #                  'CHOP', 'VHIO', 'SCI', 'PHS', 'COLU', 'UCHI']
        center_mapping_df = \
            center_mapping_df[~center_mapping_df['inputSynId'].isnull()]
        # release is a bool column
        center_mapping_df = center_mapping_df[center_mapping_df['release']]
        # center_mapping_df = center_mapping_df[
        #     ~center_mapping_df['center'].isin(exclude_sites)
        # ]
        centers = center_mapping_df.center

    if oncotree_link is None:
        onco_link = databaseToSynIdMappingDf['Id'][
            databaseToSynIdMappingDf['Database'] == 'oncotreeLink'].values[0]
        onco_link_ent = syn.get(onco_link)
        oncotree_link = onco_link_ent.externalURL
    # Check if you can connect to oncotree link,
    # if not then don't run validation / processing
    process_functions.checkUrl(oncotree_link)

    center_mapping_ent = syn.get(center_mapping_id)
    if center_mapping_ent.get('isProcessing', ['True'])[0] == 'True':
        raise Exception(
            "Processing/validation is currently happening.  "
            "Please change/add the 'isProcessing' annotation on {} "
            "to False to enable processing".format(center_mapping_id))
    else:
        center_mapping_ent.isProcessing = "True"
        center_mapping_ent = syn.store(center_mapping_ent)
    # remove this query timeout and see what happens
    # syn.table_query_timeout = 50000

    # Create new maf database, should only happen once if its specified
    if create_new_maf_database:
        today = date.today()
        table_name = f'Narrow MAF Database - {today}'
        # filetype = "vcf2maf"
        # syn7208886 is the GENIE staging project to archive maf table
        new_tables = process_functions.create_new_fileformat_table(
            syn, "vcf2maf", table_name, project_id, 'syn7208886'
        )
        syn.setPermissions(new_tables['newdb_ent'].id, 3326313, [])
        databaseToSynIdMappingDf = new_tables['newdb_mappingdf']

    format_registry = config.collect_format_types(args.format_registry_packages)
    for process_center in centers:
        input_to_database.center_input_to_database(
            syn, project_id, process_center, process,
            only_validate, databaseToSynIdMappingDf,
            center_mapping_df,
            delete_old=delete_old,
            oncotree_link=oncotree_link,
            format_registry=format_registry,
            genie_annotation_pkg=genie_annotation_pkg
        )

    # To ensure that this is the new entity
    center_mapping_ent = syn.get(center_mapping_id)
    center_mapping_ent.isProcessing = "False"
    center_mapping_ent = syn.store(center_mapping_ent)

    error_tracker_synid = process_functions.getDatabaseSynId(
        syn, "errorTracker", databaseToSynIdMappingDf=databaseToSynIdMappingDf)
    # Only write out invalid reasons if the center
    # isnt specified and if only validate
    if center is None and only_validate:
        logger.info("WRITING INVALID REASONS TO CENTER STAGING DIRS")
        write_invalid_reasons.write(syn, center_mapping_df,
                                    error_tracker_synid)
    logger.info("INPUT TO DATABASE COMPLETE")


if __name__ == "__main__":
    # Argument parsers
    # TODO: Fix case of arguments
    parser = argparse.ArgumentParser(
        description='GENIE center ')
    parser.add_argument(
        "process",
        choices=['mutation', 'main'],
        help='Process vcf, maf or the rest of the files')
    parser.add_argument(
        "--project_id",
        help="Synapse Project ID where data is stored.",
        required=True)
    parser.add_argument(
        '--center',
        help='The centers')
    parser.add_argument(
        "--pemFile",
        type=str,
        help="Path to PEM file (genie.pem)")
    parser.add_argument(
        "--deleteOld",
        action='store_true',
        help="Delete all old processed and temp files")
    parser.add_argument(
        "--onlyValidate",
        action='store_true',
        help="Only validate the files, don't process")
    parser.add_argument(
        "--oncotree_link",
        type=str,
        help="Link to oncotree code")
    parser.add_argument(
        "--createNewMafDatabase",
        action='store_true',
        help="Creates a new maf database")
    parser.add_argument(
        "--debug",
        action='store_true',
        help="Add debug mode to synapse")
    parser.add_argument(
        "--genie_annotation_pkg",
        help="GENIE annotation pkg"
    )

    # DEFAULT PARAMS
    parser.add_argument(
        "--format_registry_packages", type=str, nargs="+",
        default=["genie_registry"],
        help="Python package name(s) to get valid file formats from "
             "(default: %(default)s)."
    )
    args = parser.parse_args()

    main(args.process,
         project_id=args.project_id,
         center=args.center,
         pemfile=args.pemFile,
         delete_old=args.deleteOld,
         only_validate=args.onlyValidate,
         oncotree_link=args.oncotree_link,
         create_new_maf_database=args.createNewMafDatabase,
         debug=args.debug,
         genie_annotation_pkg=args.genie_annotation_pkg,
         format_registry=args.format_registry_packages)



from unittest import mock
from unittest.mock import patch

import pandas as pd
import pytest
import synapseclient

from genie_registry.mutationsInCis import mutationsInCis


def createMockTable(dataframe):
    table = mock.create_autospec(synapseclient.table.CsvFileTable)
    table.asDataFrame.return_value = dataframe
    return(table)


def table_query_results(*args):
    return(table_query_results_map[args])


databaseMapping = pd.DataFrame(dict(Database=['mutationsInCis'],
                                    Id=['syn7765462']))
mutCisDf = pd.DataFrame(dict(Flag=['mutationsInCis'],
                             Center=['SAGE'],
                             Tumor_Sample_Barcode=["GENIE-SAGE-ID1-1"],
                             Variant_Classification=["Nonsense_Mutation"],
                             Hugo_Symbol=["AKT1"],
                             HGVSp_Short=["p.1234"],
                             Chromosome=["1"],
                             Start_Position=[324234],
                             Reference_Allele=["AGC"],
                             Tumor_Seq_Allele2=["GCCCT"],
                             t_alt_count_num=[3],
                             t_depth=[234]))

# This is the gene positions that all bed dataframe will be processed against
table_query_results_map = {
    ("SELECT * FROM syn10967259",): createMockTable(databaseMapping),
    ("select * from syn7765462 where Center = 'SAGE'",):
        createMockTable(mutCisDf)
}

syn = mock.create_autospec(synapseclient.Synapse)
syn.tableQuery.side_effect = table_query_results

mutCis = mutationsInCis(syn, "SAGE")


ENTITY = synapseclient.Project("testing",
                               annotations={'dbMapping': ["syn10967259"]})

def test_processing():
    pass


def test_validation():
    with pytest.raises(AssertionError):
        mutCis.validateFilename(["foo"])
    assert mutCis.validateFilename(
        ["mutationsInCis_filtered_samples.csv"]) == "mutationsInCis"

    mutCisDf = pd.DataFrame(dict(Flag=['mutationsInCis'],
                                 Center=['SAGE'],
                                 Tumor_Sample_Barcode=["GENIE-SAGE-ID1-1"],
                                 Variant_Classification=["Nonsense_Mutation"],
                                 Hugo_Symbol=["AKT1"],
                                 HGVSp_Short=["p.1234"],
                                 Chromosome=["1"],
                                 Start_Position=[324234],
                                 Reference_Allele=["AGC"],
                                 Tumor_Seq_Allele2=["GCCCT"],
                                 t_alt_count_num=[3],
                                 t_depth=[234]))
    with patch.object(syn, "get", return_value=ENTITY):
        error, warning = mutCis._validate(mutCisDf, "syn1234")
        assert error == ""
        assert warning == ""
    mutCisDf = pd.DataFrame(dict(Flag=['mutationsInCis'],
                                 Center=['SAGE'],
                                 Tumor_Sample_Barcode=["GENIE-SAGE-ID1-1"],
                                 Hugo_Symbol=["AKT1"],
                                 HGVSp_Short=["p.1234"],
                                 Chromosome=["1"],
                                 Reference_Allele=["AGC"],
                                 Tumor_Seq_Allele2=["GCCCT"],
                                 t_alt_count_num=[3],
                                 t_depth=[234]))
    with patch.object(syn, "get", return_value=ENTITY):
        error, warning = mutCis._validate(mutCisDf, "syn1234")
        expectedErrors = (
            "Mutations In Cis Filter File: Must at least have these headers: "
            "Variant_Classification,Start_Position.\n")
        assert error == expectedErrors

    mutCisDf = pd.DataFrame(dict(Flag=['mutationsInCis'],
                                 Center=['SAGE'],
                                 Tumor_Sample_Barcode=["GENIE-SAGE-ID1-1"],
                                 Variant_Classification=["Nonsense_Mutation"],
                                 Hugo_Symbol=["AKT1"],
                                 HGVSp_Short=["foo"],
                                 Chromosome=["1"],
                                 Start_Position=[324234],
                                 Reference_Allele=["AGC"],
                                 Tumor_Seq_Allele2=["GCCCT"],
                                 t_alt_count_num=[3],
                                 t_depth=[234]))
    with patch.object(syn, "get", return_value=ENTITY):
        error, warning = mutCis._validate(mutCisDf, "syn1234")
        expectedErrors = (
            "Mutations In Cis Filter File: All variants must come from "
            "the original mutationInCis_filtered_samples.csv file in each "
            "institution's staging folder.\n")
        assert error == expectedErrors



from unittest import mock

import pandas as pd
import pytest
import synapseclient

from genie_registry.vcf import vcf

syn = mock.create_autospec(synapseclient.Synapse)

vcfClass = vcf(syn, "SAGE")


def test_processing():
    pass


@pytest.mark.parametrize("filepath_list", [["foo"], ["GENIE-SAGE-ID1.bed"]])
def test_incorrect_validatefilename(filepath_list):
    with pytest.raises(AssertionError):
        vcfClass.validateFilename(filepath_list)


def test_validation():
    assert vcfClass.validateFilename(["GENIE-SAGE-ID1.vcf"]) == "vcf"

    vcfDf = pd.DataFrame({
        "#CHROM": ['2', '9', '12'],
        "POS": [69688533, 99401860, 53701241],
        "ID": ['AAK1', 'AAED1', 'AAAS'],
        "REF": ['AAK1', 'AAED1', 'AAAS'],
        "ALT": ['AAK1', 'AAED1', 'AAAS'],
        "QUAL": ['AAK1', 'AAED1', 'AAAS'],
        "FILTER": ['AAK1', 'AAED1', 'AAAS'],
        "INFO": ['AAK1', 'AAED1', 'AAAS']})

    error, warning = vcfClass._validate(vcfDf)
    assert error == ""
    assert warning == ""

    vcfDf = pd.DataFrame({
        "POS": [69688533, 99401860, 53701241],
        "ID": ['AAK1', 'AAED1', 'AAAS'],
        "REF": ['AAK1', 'AAED1', 'AAAS'],
        "ALT": ['AAK1', 'AAED1', 'AAAS'],
        "QUAL": ['AAK1', 'AAED1', 'AAAS'],
        "FILTER": ['AAK1', 'AAED1', 'AAAS'],
        "INFO": ['AAK1', 'AAED1', 'AAAS'],
        "FOO": ['AAK1', 'AAED1', 'AAAS'],
        "DOO": ['AAK1', 'AA ED1', 'AAAS']})

    error, warning = vcfClass._validate(vcfDf)
    expectedError = (
        "Your vcf file must have these headers: CHROM, POS, ID, REF, "
        "ALT, QUAL, FILTER, INFO.\n"
        "Your vcf file must have FORMAT header if genotype columns exist.\n")
    expectedWarning = (
        "Your vcf file should not have any white spaces "
        "in any of the columns.\n")
    assert error == expectedError
    assert warning == expectedWarning

    vcfDf = pd.DataFrame({"#CHROM": ['chr2', 'chrM', float('nan'), 'chr2'],
                          "POS": [69688533, 99401860, 53701241, 69688533],
                          "ID": ['AAK1', 'AAED1', 'AAAS', 'AAK1'],
                          "REF": ['AAK1', 'AAED1', 'AAAS', 'AAK1'],
                          "ALT": ['AAK1', 'AAED1', 'AAAS', 'AAK1'],
                          "QUAL": ['AAK1', 'AAED1', 'AAAS', 'AAK1'],
                          "FILTER": ['AAK1', 'AAED1', 'AAAS', 'AAK1'],
                          "INFO": ['AAK1', 'AAED1', 'AAAS', 'AAK1']})

    error, warning = vcfClass._validate(vcfDf)
    expectedError = ("Your vcf file should not have duplicate rows\n"
                     "Your vcf file may contain rows that are "
                     "space delimited instead of tab delimited.\n"
                     "Your vcf file must not have variants on chrM.\n")
    expectedWarning = ("Your vcf file should not have the chr prefix "
                       "in front of chromosomes.\n")
    assert error == expectedError
    assert warning == expectedWarning



import logging
import os

import pandas as pd

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

logger = logging.getLogger(__name__)


class seg(FileTypeFormat):

    _fileType = "seg"

    _process_kwargs = ["newPath", "databaseSynId"]

    def _validateFilename(self, filePath):
        assert os.path.basename(filePath[0]) == "genie_data_cna_hg19_%s.%s" % (self.center, self._fileType)

    def _process(self, seg):
        seg.columns = [col.upper() for col in seg.columns]
        newsamples = [process_functions.checkGenieId(i, self.center) for i in seg['ID']]
        seg['ID'] = newsamples
        seg = seg.drop_duplicates()
        seg = seg.rename(columns= {'LOC.START':'LOCSTART','LOC.END':'LOCEND','SEG.MEAN':'SEGMEAN','NUM.MARK':'NUMMARK'})
        seg['CHROM'] = [str(chrom).replace("chr","") for chrom in seg['CHROM']]
        seg['CENTER'] = self.center
        seg['LOCSTART'] = seg['LOCSTART'].astype(int)
        seg['LOCEND'] = seg['LOCEND'].astype(int)
        seg['NUMMARK'] = seg['NUMMARK'].astype(int)
        return(seg)

    def process_steps(self, seg, newPath, databaseSynId):
        seg = self._process(seg)
        process_functions.updateData(self.syn, databaseSynId, seg, self.center, toDelete=True)
        seg.to_csv(newPath,sep="\t",index=False)
        return(newPath)

    def _validate(self, segDF):
        total_error = ""
        warning = ""
        segDF.columns = [col.upper() for col in segDF.columns]

        REQUIRED_HEADERS = pd.Series(['ID','CHROM','LOC.START','LOC.END','NUM.MARK','SEG.MEAN'])
        
        if not all(REQUIRED_HEADERS.isin(segDF.columns)):
            total_error += "Your seg file is missing these headers: %s.\n" % ", ".join(REQUIRED_HEADERS[~REQUIRED_HEADERS.isin(segDF.columns)])
        else:
            intCols = ['LOC.START','LOC.END','NUM.MARK']
            nonInts = [col for col in intCols if segDF[col].dtype != int]
            if len(nonInts) > 0:
                total_error += "Seg: Only integars allowed in these column(s): %s.\n" % ", ".join(sorted(nonInts))
            if not segDF['SEG.MEAN'].dtype in [float, int]:
                total_error += "Seg: Only numerical values allowed in SEG.MEAN.\n"

        checkNA = segDF.isna().apply(sum)
        nullCols = [ind for ind in checkNA.index if checkNA[ind] > 0]
        if len(nullCols) > 0:
            total_error += "Seg: No null or empty values allowed in column(s): %s.\n" % ", ".join(sorted(nullCols))

        return(total_error, warning)



import logging
import os

import pandas as pd

logger = logging.getLogger(__name__)


class FileTypeFormat(object):

    _process_kwargs = ["newPath", "databaseSynId"]

    _fileType = "fileType"

    _validation_kwargs = []

    def __init__(self, syn, center, poolSize=1):
        self.syn = syn
        self.center = center

        # self.pool = multiprocessing.Pool(poolSize)

    def _get_dataframe(self, filePathList):
        '''
        This function by defaults assumes the filePathList is length of 1
        and is a tsv file.  Could change depending on file type.

        Args:
            filePathList:  A list of file paths (Max is 2 for the two
                           clinical files)

        Returns:
            df: Pandas dataframe of file
        '''
        filePath = filePathList[0]
        df = pd.read_csv(filePath, sep="\t", comment="#")
        return(df)

    def read_file(self, filePathList):
        '''
        Each file is to be read in for validation and processing.
        This is not to be changed in any functions.

        Args:
            filePathList:  A list of file paths (Max is 2 for the two
                           clinical files)

        Returns:
            df: Pandas dataframe of file
        '''
        df = self._get_dataframe(filePathList)
        return(df)

    def _validateFilename(self, filePath):
        '''
        Function that changes per file type for validating its filename
        Expects an assertion error.

        Args:
            filePath: Path to file
        '''
        # assert True
        raise NotImplementedError

    def validateFilename(self, filePath):
        '''
        Validation of file name.  The filename is what maps the file
        to its validation and processing.

        Args:
            filePath: Path to file

        Returns:
            str: file type defined by self._fileType
        '''
        self._validateFilename(filePath)
        return(self._fileType)

    def process_steps(self, df, **kwargs):
        '''
        This function is modified for every single file.
        It reformats the file and stores the file into database and Synapse.
        '''
        pass

    def preprocess(self, newpath):
        '''
        This is for any preprocessing that has to occur to the entity name
        to add to kwargs for processing.  entity name is included in
        the new path

        Args:
            newpath: Path to file
        '''
        return dict()

    def process(self, filePath, **kwargs):
        '''
        This is the main processing function.

        Args:
            filePath: Path to file
            kwargs: The kwargs are determined by self._process_kwargs

        Returns:
            str: file path of processed file
        '''
        preprocess_args = self.preprocess(kwargs.get('newPath'))
        kwargs.update(preprocess_args)
        mykwargs = {}
        for required_parameter in self._process_kwargs:
            assert required_parameter in kwargs.keys(), \
                "%s not in parameter list" % required_parameter
            mykwargs[required_parameter] = kwargs[required_parameter]
        logger.info('PROCESSING %s' % filePath)
        # This is done because the clinical files are being merged into a list
        if self._fileType == "clinical":
            path_or_df = self.read_file(filePath)
        # If file type is vcf or maf file, processing requires a filepath
        elif self._fileType not in ['vcf', 'maf', 'mafSP', 'md']:
            path_or_df = self.read_file([filePath])
        else:
            path_or_df = filePath
        path = self.process_steps(path_or_df, **mykwargs)
        return(path)

    def _validate(self, df, **kwargs):
        '''
        This is the base validation function.
        By default, no validation occurs.

        Args:
            df: A dataframe of the file
            kwargs: The kwargs are determined by self._validation_kwargs

        Returns:
            tuple: The errors and warnings as a file from validation.
                   Defaults to blank strings
        '''
        errors = ""
        warnings = ""
        logger.info("NO VALIDATION for %s files" % self._fileType)
        return(errors, warnings)

    def validate(self, filePathList, **kwargs):
        '''
        This is the main validation function.
        Every file type calls self._validate, which is different.

        Args:
            filePathList: A list of file paths.
            kwargs: The kwargs are determined by self._validation_kwargs

        Returns:
            tuple: The errors and warnings as a file from validation.
        '''
        mykwargs = {}
        for required_parameter in self._validation_kwargs:
            assert required_parameter in kwargs.keys(), "%s not in parameter list" % required_parameter
            mykwargs[required_parameter] = kwargs[required_parameter]

        errors = ""

        try:
            df = self.read_file(filePathList)
        except Exception as e:
            errors = "The file(s) ({filePathList}) cannot be read. Original error: {exception}".format(filePathList=filePathList,
                                                                                                       exception=str(e))
            warnings = ""

        if not errors:
            logger.info("VALIDATING %s" % os.path.basename(",".join(filePathList)))
            errors, warnings = self._validate(df, **mykwargs)
        
        valid = (errors == '')
        
        return valid, errors, warnings


import os
import logging

import pandas as pd
from genie import process_functions

from .maf import maf

logger = logging.getLogger(__name__)


class mafSP(maf):
    '''
    MAF SP file format validation / processing
    '''

    _fileType = "mafSP"

    def _validateFilename(self, filePath):
        '''
        Validates filename.  Should be
        nonGENIE_data_mutations_extended_CENTER.txt
        '''
        assert os.path.basename(filePath[0]) == \
            "nonGENIE_data_mutations_extended_{}.txt".format(self.center)

    def formatMAF(self, mafDf):
        '''
        The sponsored project maf file doesn't have less columns
        '''
        mafDf['Center'] = self.center
        mafDf['Tumor_Sample_Barcode'] = [
            process_functions.checkGenieId(i, self.center)
            for i in mafDf['Tumor_Sample_Barcode']]
        mafDf['Sequence_Source'] = float('nan')
        mafDf['Sequencer'] = float('nan')
        mafDf['Validation_Status'][
            mafDf['Validation_Status'].isin(["Unknown", "unknown"])] = ''
        return(mafDf)

    def storeProcessedMaf(
            self, filePath, mafSynId, centerMafSynId, isNarrow=False):
        '''
        Stores SP maf to database
        '''
        logger.info('STORING %s' % filePath)
        mafDataFrame = pd.read_csv(filePath, sep="\t")
        process_functions.updateData(
            self.syn, mafSynId, mafDataFrame,
            self.center, filterByColumn="Center", toDelete=True)
        return(filePath)



from unittest import mock
from unittest.mock import Mock, patch
import uuid

import pandas as pd
import pytest
import synapseclient

from genie import process_functions

syn = mock.create_autospec(synapseclient.Synapse)

DATABASE_DF = pd.DataFrame({
    'UNIQUE_KEY': ['test1', 'test2', 'test3'],
    "test": ['test1', 'test2', 'test3'],
    "foo": [1, 2, 3],
    "baz": [float('nan'), float('nan'), float('nan')]})
DATABASE_DF.index = ['1_3', '2_3', '3_5']
ENTITY = synapseclient.Project("foo", annotations={"dbMapping": ["syn1234"]})

@pytest.mark.parametrize("input_str,output", [
        ("1.0\t", "1\t"),
        ("1.0\n", "1\n"),
        ("1.5\t", "1.5\t"),
        ("1\t", "1\t"),
        ("0\t", "0\t"),
        ("'a'\t'b'\n1.0\t2.0\n", "'a'\t'b'\n1\t2\n"),
    ])
def test_removeStringFloat(input_str, output):
    """Remove string float - will always assume that there is a \n
    at the end.  This is because if a value was 2.01, we dont want to
    remove the .0 from this."""
    assert process_functions.removeStringFloat(input_str) == output


def test_valid__check_valid_df():
    process_functions._check_valid_df(DATABASE_DF, "test")


def test_invalid__check_valid_df():
    with pytest.raises(ValueError, match="Must pass in pandas dataframe"):
        process_functions._check_valid_df("foo", "test")
    with pytest.raises(
            ValueError,
            match="'error' column must exist in dataframe"):
        process_functions._check_valid_df(DATABASE_DF, "error")


def test__get_left_diff_df():
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test1', 'test2', 'test3', 'test4'],
        "test": ['test1', 'test2', 'test3', 'test4'],
        "foo": [1, 2, 3, 4],
        "baz": [float('nan'), float('nan'), float('nan'), 3.2]})
    get_diff = process_functions._get_left_diff_df(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    expecteddf = new_datadf.loc[[3]]
    assert get_diff.equals(expecteddf[get_diff.columns])


def test_norows_get_left_diff_df():
    append_rows = process_functions._get_left_diff_df(
        DATABASE_DF, DATABASE_DF, 'UNIQUE_KEY')
    assert append_rows.empty


def test_first_validation_get_left_diff_df():
    '''
    This checks to make sure that validation is called
    - In a situation where someone comments out the validation
    line, this will cause an error
    '''
    with pytest.raises(
            ValueError,
            match="'FOO' column must exist in dataframe"):
        process_functions._get_left_diff_df(
            DATABASE_DF, DATABASE_DF, 'FOO')


def test_second_validation_get_left_diff_df():
    '''
    This checks to make sure that validation is called
    - In a situation where someone comments out the validation
    line, this will cause an error
    '''
    testing = DATABASE_DF.copy()
    testing['FOO'] = float('nan')
    with pytest.raises(
            ValueError,
            match="'FOO' column must exist in dataframe"):
        process_functions._get_left_diff_df(
            testing, DATABASE_DF, 'FOO')


def test_first_validation_get_left_union_df():
    '''
    This checks to make sure that validation is called
    - In a situation where someone comments out the 1st validation
    line, this will cause an error
    '''
    with pytest.raises(
            ValueError,
            match="'FOO' column must exist in dataframe"):
        process_functions._get_left_union_df(
            DATABASE_DF, DATABASE_DF, 'FOO')


def test_second_validation_get_left_union_df():
    '''
    This checks to make sure that validation is called
    - In a situation where someone comments out the 2nd validation
    line, this will cause an error
    '''
    testing = DATABASE_DF.copy()
    testing['FOO'] = float('nan')
    with pytest.raises(
            ValueError,
            match="'FOO' column must exist in dataframe"):
        process_functions._get_left_union_df(
            testing, DATABASE_DF, 'FOO')


def test_append__append_rows():
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test1', 'test2', 'test3', 'test4'],
        "test": ['test1', 'test2', 'test3', 'test4'],
        "foo": [1, 2, 3, 4],
        "baz": [float('nan'), float('nan'), float('nan'), 3.2]})
    expecteddf = pd.DataFrame({
        'test': ['test4'],
        'foo': [4],
        'baz': [3.2]})
    append_rows = process_functions._append_rows(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    append_rows.fillna('', inplace=True)
    expecteddf.fillna('', inplace=True)
    assert append_rows.equals(expecteddf[append_rows.columns])


def test___create_update_rowsdf():
    differentrows = [True, True, False]
    database = pd.DataFrame({"test": ['test', 'test2', 'test3'],
                             "foo": [1, 3, 3],
                             "baz": [float('nan'), 5, float('nan')]},
                            index=['test1', 'test5', 'test4'])
    new_datadf = pd.DataFrame({"test": ['test1', 'test4', 'test3'],
                               "foo": [2, 3, 3],
                               "baz": [3, 5, float('nan')]},
                              index=['test1', 'test5', 'test4'])

    to_update_rowsdf = process_functions._create_update_rowsdf(
        database, new_datadf, DATABASE_DF.index, differentrows)
    expecteddf = pd.DataFrame({
        "test": ['test1', 'test4'],
        "foo": [2, 3],
        "baz": [3.0, 5.0],
        "ROW_ID": ["1", "2"],
        "ROW_VERSION": ["3", "3"]})
    assert to_update_rowsdf.equals(expecteddf[to_update_rowsdf.columns])


def test_none__create_update_rowsdf():
    differentrows = [False, False, False]
    database = pd.DataFrame({
        "test": ['test', 'test2', 'test3'],
        "foo": [1, 3, 3],
        "baz": [float('nan'), 5, float('nan')]},
        index=['test1', 'test5', 'test4'])
    new_datadf = pd.DataFrame({
        "test": ['test1', 'test4', 'test3'],
        "foo": [2, 3, 3],
        "baz": [3, 5, float('nan')]},
        index=['test1', 'test5', 'test4'])

    to_update_rowsdf = process_functions._create_update_rowsdf(
        database, new_datadf, DATABASE_DF.index, differentrows)
    assert to_update_rowsdf.empty


def test___get_left_union_df():
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test1', 'test5', 'test4'],
        "test": ['test', 'test2', 'test3'],
        "foo": [1, 3, 3],
        "baz": [float('nan'), 5, float('nan')]})
    left_union = process_functions._get_left_union_df(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    expecteddf = pd.DataFrame({
        'UNIQUE_KEY': ['test1'],
        'test': ['test'],
        'foo': [1],
        'baz': [float('nan')]})
    assert left_union.equals(expecteddf[left_union.columns])


def test_none__get_left_union_df():
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test7', 'test5', 'test4'],
        "test": ['test', 'test2', 'test3'],
        "foo": [1, 3, 3],
        "baz": [float('nan'), 5, float('nan')]})
    left_union = process_functions._get_left_union_df(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    assert left_union.empty


def test_update__update_rows():
    '''
    Tests index comparison for updating rows
    '''
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test1', 'test2', 'test3'],
        "test": ['test', 'test2', 'test3'],
        "foo": [1, 3, 3],
        "baz": [float('nan'), 5, float('nan')]})

    expecteddf = pd.DataFrame({
        "test": ['test', 'test2'],
        "foo": [1, 3],
        "baz": ['', 5],
        'ROW_ID': ['1', '2'],
        'ROW_VERSION': ['3', '3']})
    update_rows = process_functions._update_rows(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    assert update_rows.equals(expecteddf[update_rows.columns])


def test_maintaintype__update_rows():
    '''
    Test pandas behavior.  Integer -> Float if NA exists
    '''
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test1', 'test2', 'test3'],
        "test": ['test1', 'test2', 'test3'],
        "foo": [1, 3, 3],
        "baz": [float('nan'), 5, float('nan')]})
    # Test that the datatype passed into from new_datadf gets preserved
    expecteddf = pd.DataFrame({
        "test": ['test2'],
        "foo": [3],
        "baz": [5],
        'ROW_ID': ['2'],
        'ROW_VERSION': ['3']})
    expecteddf = expecteddf.astype({'baz': object})
    update_rows = process_functions._update_rows(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    assert update_rows.equals(expecteddf[update_rows.columns])


def test_noupdate__update_rows():
    '''
    Tests the index comparison to get no updates
    '''
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test4'],
        "test": ['test'],
        "foo": [1],
        "baz": [float('nan')]})
    update_rows = process_functions._update_rows(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    assert update_rows.empty


def test_delete__delete_rows():
    new_datadf = pd.DataFrame({
        'UNIQUE_KEY': ['test1'],
        "test": ['test1'],
        "foo": [1],
        "baz": [float('nan')]})
    expecteddf = pd.DataFrame({
        0: ['2', '3'],
        1: ['3', '5']})
    delete_rows = process_functions._delete_rows(
        new_datadf, DATABASE_DF, 'UNIQUE_KEY')
    assert delete_rows.equals(expecteddf)


def test_norows__delete_rows():
    delete_rows = process_functions._delete_rows(
        DATABASE_DF, DATABASE_DF, 'UNIQUE_KEY')
    assert delete_rows.empty


class argparser:
    def asDataFrame(self):
        database_dict = {"Database": ["centerMapping"],
                         "Id": ["syn123"]}
        databasetosynid_mappingdf = pd.DataFrame(database_dict)
        return(databasetosynid_mappingdf)


@pytest.mark.parametrize("test,staging,synid", [(False, False, "syn10967259"),
                                                (False, True, "syn12094210"),
                                                (True, False, "syn11600968")])
def test_get_synid_database_mappingdf(test, staging, synid):
    '''
    Tests getting database mapping config
    no flags
    staging flag
    test flag
    '''
    arg = argparser()
    with patch.object(syn, "get", return_value=ENTITY), \
         patch.object(process_functions, "get_syntabledf",
                      return_value=arg.asDataFrame()) as patch_gettabledf:
        df = process_functions.get_synid_database_mappingdf(
            syn, project_id=None)
        patch_gettabledf.assert_called_once_with(
            syn, "SELECT * FROM {}".format(ENTITY.dbMapping[0]))
        assert df.equals(arg.asDataFrame())


def test_get_syntabledf():
    '''
    Test helper function that queries synapse tables and returns dataframes
    '''
    arg = argparser()
    with patch.object(syn, "tableQuery",
                      return_value=arg) as patch_syn_tablequery:
        querystring = "select * from foo"
        df = process_functions.get_syntabledf(syn, querystring)
        patch_syn_tablequery.assert_called_once_with(querystring)
        assert df.equals(arg.asDataFrame())


def test__create_schema():
    """Tests calling of create schema"""
    table_name = str(uuid.uuid1())
    parentid = str(uuid.uuid1())
    columns = [str(uuid.uuid1())]
    annotations = {"foo": "bar"}

    schema = synapseclient.Schema(table_name, columns=columns,
                                  parent=parentid, annotations=annotations)
    with patch.object(syn, "store",
                      return_value=schema) as patch_syn_store:

        new_schema = process_functions._create_schema(syn, table_name, parentid,
                                                      columns=columns,
                                                      annotations=annotations)
        patch_syn_store.assert_called_once_with(schema)
        assert new_schema == schema


def test__update_database_mapping():
    """Tests updates database mapping"""
    fileformat = str(uuid.uuid1())
    database_mappingdf = pd.DataFrame({'Database': [fileformat, "foo"],
                                       "Id": ['11111', "bar"]})
    database_mapping_synid = str(uuid.uuid1())
    new_tableid = str(uuid.uuid1())
    expected_mapdf = pd.DataFrame({'Database': [fileformat, "foo"],
                                   "Id": [new_tableid, "bar"]})
    with patch.object(syn, "store") as patch_syn_store:
        newdb = process_functions._update_database_mapping(syn, database_mappingdf,
                                                           database_mapping_synid,
                                                           fileformat, new_tableid)
        assert newdb.equals(expected_mapdf)
        patch_syn_store.assert_called_once()


def test_noname__move_entity():
    """Tests not changing entity name"""
    ent = synapseclient.Entity(name="foo", parentId="syn2222")
    new_parent = "syn1234"
    with patch.object(syn, "store") as patch_syn_store:
        process_functions._move_entity(syn, ent, new_parent)
        ent.parentId = new_parent
        patch_syn_store.assert_called_once_with(ent)


def test_name__move_entity():
    """Tests entity name is updated"""
    ent = synapseclient.Entity(name="foo", parentId="syn2222")
    new_parent = "syn1234"
    new_name = "updated name"
    with patch.object(syn, "store") as patch_syn_store:
        process_functions._move_entity(syn, ent, new_parent, new_name)
        ent.parentId = new_parent
        ent.name = new_name
        patch_syn_store.assert_called_once_with(ent)


def test_create_new_fileformat_table():
    fileformat = str(uuid.uuid1())
    db_synid = "syn1111111"
    database_mappingdf = pd.DataFrame({'Database': [fileformat, "foo"],
                                       "Id": [db_synid, "bar"]})
    db_mapping_info = {'synid': 'syn666',
                       'df': database_mappingdf}
    table_ent = synapseclient.Entity(
        parentId="syn123", name="foo", primaryKey=['annot'], id='syn12345'
    )
    project_id = "syn234"
    archived_project_id = "syn23333"
    new_table_name = str(uuid.uuid1())

    new_table_ent = synapseclient.Entity(
        parentId="syn123323", name="foofoo", id='syn23231'
    )
    update_return = Mock()
    move_entity_return = Mock()
    with patch.object(process_functions, "get_dbmapping",
                      return_value=db_mapping_info) as patch_getdb,\
         patch.object(syn, "get",
                      return_value=table_ent) as patch_syn_get,\
         patch.object(syn, "getTableColumns",
                      return_value=['foo', 'ddooo']) as patch_get_table_cols,\
         patch.object(process_functions, "_create_schema",
                      return_value=new_table_ent) as patch_create_schema,\
         patch.object(process_functions,
                      "_update_database_mapping",
                      return_value=update_return) as patch_update,\
         patch.object(process_functions, "_move_entity",
                      return_value=move_entity_return) as patch_move,\
         patch.object(process_functions.time, "time", return_value=2):
        new_table = process_functions.create_new_fileformat_table(
            syn, fileformat, new_table_name, project_id, archived_project_id
        )
        patch_getdb.assert_called_once_with(syn, project_id)
        patch_syn_get.assert_called_once_with(db_synid)
        patch_get_table_cols.assert_called_once_with(db_synid)
        patch_create_schema.assert_called_once_with(
            syn, table_name=new_table_name, columns=['foo', 'ddooo'],
            parentid=project_id, annotations=table_ent.annotations
        )
        patch_update.assert_called_once_with(syn, database_mappingdf,
                                             'syn666', fileformat,
                                             new_table_ent.id)
        patch_move.assert_called_once_with(syn, table_ent,
                                           archived_project_id,
                                           name="ARCHIVED 2-foo")
        assert new_table == {"newdb_ent": new_table_ent,
                             "newdb_mappingdf": update_return,
                             "moved_ent": move_entity_return}
